<?php

/**
 * @file
 * Provides drush commands to set up a site for Acquia Site Factory.
 */

/**
 * Implements hook_drush_command().
 */
function drunkins_drush_command() {
  return array(
    'drunkins-reset-job' => array(
      'description' => dt('Resets a running drinkins job and deletes all items that are still queued.

A batch process\' items cannot be deleted with this command. The job will be \'unlocked\' but if the batch process is currently still being run by another PHP process, this has unknown consequences.'),
      'arguments' => array(
        'job-id' => dt('The ID (machine name) of the job.'),
      ),
    ),
    'drunkins-run-job' => array(
      'description' => dt('Runs a full drunkins job (starting, processing all items and finishing it).'),
      'arguments' => array(
        'job-id' => dt('The ID (machine name) of the job.'),
      ),
      'required-arguments' => TRUE,      'options' => array(
        'settings-json' => dt('Settings to override the default ones for this job, in JSON format.'),
        'runner-settings-json' => dt('Runner settings to override the default ones for this job run, in JSON format. ("progressive" is ignored.)'),
      ),
    ),
    'drunkins-start-job' => array(
      'description' => dt('Starts a drunkins job in queue mode. (Other mechanisms, e.g. cron, will be responsible for running it.)

If the standard job configuration does not make it run on cron, \'{"force_cron":true}\' can be passed in --runner-settings-json.'),
      'arguments' => array(
        'job-id' => dt('The ID (machine name) of the job.'),
      ),
      'required-arguments' => TRUE,
      'options' => array(
        'settings-json' => dt('Settings to override the default ones for this job, in JSON format.'),
        'runner-settings-json' => dt('Runner settings to override the default ones for this job run, in JSON format.'),
      ),
    ),
  );
}

/**
 * Command callback: resets a Drunkins job.
 */
function drush_drunkins_reset_job($job_id) {
  drunkins_reset_job($job_id);
}

/**
 * Command callback: fully runs a Drunkins job.
 */
function drush_drunkins_run_job($job_id) {
  return drush_drunkins_start_job($job_id, array('progressive' => FALSE));
}

/**
 * Command callback: starts a Drunkins job.
 *
 * @param string $job_id
 *   The job ID.
 * @param array $override_runner_settings
 *   (optional) Runner settings that override even the drush option. If set,
 *   the job is run in 'drush' mode; if not, in 'queue' mode.
 * @return bool
 */
function drush_drunkins_start_job($job_id, array $override_runner_settings = NULL) {
  $settings = drush_get_option('settings-json', array());
  if ($settings) {
    $settings = json_decode($settings, TRUE);
    if ($settings === NULL) {
      return drush_set_error(dt('The settings-json value does not contain valid JSON.'));
    }
  }

  $runner_settings = drush_get_option('runner-settings-json', array());
  if ($runner_settings) {
    $runner_settings = json_decode($runner_settings, TRUE);
    if ($runner_settings === NULL) {
      return drush_set_error(dt('The settings-json value does not contain valid JSON.'));
    }
  }
  if (isset($override_runner_settings)) {
    $runner_settings = $override_runner_settings + $runner_settings;
  }

  // We always want to see "There are no items to process." so override any
  // log severity for that message.
  $runner_settings['log_severity_no_items'] = isset($runner_settings['log_level']) ? $runner_settings['log_level'] : WATCHDOG_INFO;

  // Change logging settings. 'screen' means drupal_set_message so remove that.
  if (isset($runner_settings['logging'])) {
    if ($runner_settings['logging'] & DrunkinsJob::LOG_SCREEN) {
      $runner_settings['logging'] -= DrunkinsJob::LOG_SCREEN;
    }
    $runner_settings['logging'] |= DrunkinsJob::LOG_DRUSH;
  }
  else {
    $runner_settings['logging'] = DrunkinsJob::LOG_DRUSH | DrunkinsJob::LOG_WATCHDOG;
  }

  // As documented in function comments (at least for now):
  $method = isset($override_runner_settings) ? 'drush' : 'queue';

// @TODO if drush "DEBUG" flag set, then also set for drunkins?

  $ret = drunkins_start_job($job_id, $method, $runner_settings, $settings);

  if ($ret && $method === 'queue') {
    // For the remainder of the process, logging should be different.
    // The process state, including runner settings, have not been saved
    // into the database yet, so assume that will get done automatically.
    $runner_settings['logging'] -= DrunkinsJob::LOG_DRUSH;
    _drunkins_set_process_config($job_id, 'runner', $runner_settings);
  }

  return $ret;
}
