<?php

/**
 * @file
 * Admin & processing (batch/queue) related functionality.
 */

// Admin / UI functions

function drunkins_overview() {

  // Note: (batch) job id == queue name.

  // TODO: check where t() is needed

  $header = array('Queue / Job', 'Type', 'Status', '# items left', 'Last start', 'Last finish', 'Operations');
  $rows = array();

  // Compile info from two sources which do not fully overlap:
  // - jobs info (including status/run info)
  // - all queues from hook_queue_info (not only the Drunkins job queues).
  // (In D7 there's probably no difference in 'all queues'. because hardly
  // anything implements hook_queue_info. Still, we will accommodate for it.)
  $jobs_info = drunkins_job_status_info();
  $queues = module_invoke_all('queue_info');
  drupal_alter('queue_info', $queues);
  foreach ($queues as $queue_name => $queue_info) {

    $runner_settings = _drunkins_get_process_config($queue_name);
    $queue = DrupalQueue::get($queue_name);
    $count = $queue->numberOfItems();
    $countinfo = $count;
    $title = empty($queue_info['title']) ? $queue_name : ($queue_info['title'] . ' (<em>' . $queue_name . '</em>)');

    if (isset($queue_info['drunkins'])) {
      $options = array();
      if (!isset($queue_info['drunkins']['cron_schedule'])) {
        $options[] = 'cron disabled';
      }
      $type = 'Drunkins'
              . ($options ? ' (' . implode('; ', $options) . ')' : '');
      if ($queue_info['drunkins']['running'] < 2) {
        $status = empty($queue_info['drunkins']['started']) ? '-' : 'finished';
        if ($queue_info['drunkins']['running'] == 1) {
          $status .= '<br/><em>(cron due'
                     . (!empty($queue_info['drunkins']['run_next'])
              ? ' since ' . format_date($queue_info['drunkins']['run_next'], 'short') : '')
                     . ')</em>';
        }
        // finished => there should be no queued items
        if ($count) {
          $countinfo .= ' <strong>(?!)</strong>';
        }
      }
      else {
        $status_detail = array($queue_info['drunkins']['method']);

        $context = isset($jobs_info[$queue_name]['process_context']) ? $jobs_info[$queue_name]['process_context'] : array();
        if (!empty($context['drunkins_process_more'])) {
          $start_runs = empty($context['drunkins_rerun_start']) ? '?' : $context['drunkins_rerun_start'];
          $start_count = isset($context['drunkins_rerun_start_count']) ? $context['drunkins_rerun_start_count'] : '?';
          $status_detail[] = "fetching more after {$start_runs} runs/{$start_count} items";
        }

        if ($queue_info['drunkins']['running'] == 3) {
          if ($queue_info['drunkins']['method'] === 'batch') {
            // There should be no queued items for a 'batch' queue job.
            // (They should be queued under another name.)
            if ($count) {
              $countinfo .= ' <strong>(?!)</strong>';
            }
            else {
              $countinfo = '?';
            }
          }
          else {
            if (!empty($runner_settings['force_cron'])) {
              $status_detail[] = 'cron forced for the current process';
            }
            if (!$count && empty($context['drunkins_process_more'])) {
              $status_detail[] = 'waiting to finish';
            }
          }
        }
        else {
          if ($queue_info['drunkins']['run_info'] == 4) {
            // It's debatable whether we should call this "running" in the UI?
            $status_detail[] = 'error on finish';
          }
          else {
            $status_detail[] = 'locked; cause unknown';
          }
        }

        $status = 'running (' . implode($status_detail, ', ') . ')';

        if (!empty($queue_info['drunkins']['cron_schedule'])) {
          // We have not calculated run_next yet, but should display it.
          _drunkins_calculate_dynamic_info($queue_info['drunkins'], $queue_name, TRUE);
        }
      }
      // Display next cron run, unless we already have.
      if ($queue_info['drunkins']['running'] != 1
          && !empty($queue_info['drunkins']['run_next'])) {
        $next = format_date($queue_info['drunkins']['run_next'], 'short');

        $status .= $queue_info['drunkins']['run_next'] >= REQUEST_TIME
          ? "<br/><em>(cron due $next)</em>" : "<br/><em>(cron due since $next)</em>";
      }

      $started = empty($queue_info['drunkins']['started']) ? '-' : format_date($queue_info['drunkins']['started']);
      $finished = empty($queue_info['drunkins']['finished']) ? '-' : format_date($queue_info['drunkins']['finished']);
      if ($finished != '-') {
        if ($queue_info['drunkins']['started'] <= $queue_info['drunkins']['finished']) {
          // Output info about previous process.
          $finished .= ' (' . $queue_info['drunkins']['method'];
          switch ($queue_info['drunkins']['run_info']) {
            case 3:
              // This very likely won't happen; {finished} is not updated on
              // "error_on_finish" in _drunkins_finish_process(). We're just
              // including it to be future proof and to not trigger the default.
              $finished .= '; error on finish)';
              break;
            case 4:
              $finished .= '; terminated on start)';
              break;
            case 5:
              $finished .= '; reset manually)';
              break;
            case 0:
              $finished .= ')';
              break;
            default:
              // Other states always mean 'running', so something's wrong:
              // finished should be smaller than started.
            $finished .= ' (<em>data error</em>)';
          }
        }
        else {
          // 'Last finish' is not connected with 'last start'. Do we need to specify this?
          $finished = '<em>(' . $finished . ')</em>';
        }
      }
      $actions = array();
      if ($queue_info['drunkins']['running'] < 2) {
        $actions[] = l('Start job / view data', "admin/config/workflow/drunkins/$queue_name");
  //      $actions[] = l('Enqueue (for cron processing)', "admin/config/workflow/drunkins/$queue_name/...");
      }
      elseif (user_access('access drunkins advanced')) {
        $comment = $count ? 'delete items & ' : '';
        $actions[] = l('Reset job (' . $comment . 'remove lock)', "admin/config/workflow/drunkins/$queue_name/reset");
  //      if ($queue_info['running'] == 3) {
  //        $actions[] = l('Remove leases', "admin/config/workflow/drunkins/$queue_name/...");
  //      }
      }
      // @todo for 'waiting to finish' job, make link to finish it.
    }
    else {
      // Non-Drunkins queue
      // Check what we know from looking at queue_ui code
      $type = (isset($queue_info['batch']) || isset($queue_info['cron']['callback']))
        ? 'queue_ui' : '??';
      $status = $started = $finished = '?';
      // We can move queue_ui actions here, if we want... see also commented code above.
      $actions = array();
    }

    $rows[] = array($title, $type, $status, $countinfo, $started, $finished,
      implode(' - ', $actions)
    );
  }

  // Get queue table info for later.
  $other_queues = db_query("SELECT name, count(created) AS cnt, min(created) as crt FROM {queue} GROUP BY name")->fetchAllAssoc('name', PDO::FETCH_ASSOC);

  // D7 core still has separate cron queues. Our Drunkins jobs also have them,
  // so don't process these again. Since we're not sure if a queue uses the
  // {queue} table, we get the count through the official method. For 'started'
  // there is no official method, so we use the above if we have it.
  $cron_queues = module_invoke_all('cron_queue_info');
  drupal_alter('queue_info', $cron_queues);
  $cron_queues = array_diff_key($cron_queues, $queues);
  foreach ($cron_queues as $queue_name => $queue_info) {
    $title = empty($queue_info['title']) ? $queue_name : ($queue_info['title'] . ' (<em>' . $queue_name . '</em>)');
    $count = DrupalQueue::get($queue_name)->numberOfItems();
    $rows[] = array($title, 'Cron queue', $count ? 'running' : 'finished', $count,
      isset($other_queues[$queue_name]['crt']) ? format_date($other_queues[$queue_name]['crt']) : '?',
      '?', '');
    // TODO We can probably reset/remove leases here...
  }

  // For fun, see what other queues we have in the {queue} table.
  // Instead of executing DrupalQueue::get($queue_name)->numberOfItems() for
  // each queue, we might as well get the count from the table immediately.
  $other_queues = array_diff_key($other_queues, $queues, $cron_queues);
  foreach ($other_queues as $queue_name => $queue_info) {
    $rows[] = array($queue_name, 'queue table', $queue_info['cnt'] ? 'running' : 'finished',
      $queue_info['cnt'], format_date($queue_info['crt']), '?', '');
    // TODO We can probably reset/remove leases here...
  }

  return array(
    '#theme' => 'table',
    '#header' => $header,
    '#rows' => $rows,
  );
}

/**
 * Title callback for various administrative screens.
 * @see drunkins_menu()
 */
function drunkins_title_callback($title, $job_id) {
  $job_info = drunkins_job_status_info($job_id);
  return str_replace(
    '%job_id',
    "'" . (isset($job_info['title']) ? $job_info['title'] : $job_id) . "'",
    $title);
}

/**
 * Form definition: reset a running proces.
 */
function drunkins_job_reset_form($form, &$form_state, $job_id) {
  $job_info = drunkins_job_status_info($job_id);

  drupal_set_title(t("Reset job %job", array('%job' => $job_id) ));

  $comment = '';
  if ($job_info['method'] === 'batch') {
    // The queued items for the batch process are not in our queue. They are
    // in a 'batch:NNN' queue, where we do not know NNN. See _batch_queue().

    // Only set warning if the form was not just submitted.
    if (empty($form_state['input'])) {
      drupal_set_message(t("Please note that a batch process' items cannot be deleted through here. The job will be 'unlocked' but if the batch process is currently still being run by another client, this has unknown consequences."), 'warning');
    }
  }
  else {
    $comment = ' and delete @cnt items';
  }

  // Output the same link as currently done in the full settings/view screen,
  // mainly because we don't have the link to that screen on the UI overview for
  // running processes.
  // @todo Fix the latter?
  $form['display_process_data'] = array(
    '#markup' => l(t('Display metadata used by the current process.'), "admin/config/workflow/drunkins/$job_id/metadata"),
    '#weight' => 200,
    '#prefix' => '<div class="description">',
    '#suffix' => '</div>',
  );
  // And items. (Maybe fix up later, show in UI overview too?)
  $form['display_process_items'] = array(
    '#markup' => l(t('Display items queued by the current process.'), "admin/config/workflow/drunkins/$job_id/items"),
    '#weight' => 201,
    '#prefix' => '<div class="description">',
    '#suffix' => '</div>',
  );

  $form['job_id'] = array('#type' => 'value', '#value' => $job_id);
  return confirm_form($form,
    t('Are you sure you want to delete the running job ' . $comment . '?', array('@cnt' => DrupalQueue::get($job_id)->numberOfItems())),
    'admin/config/workflow/drunkins'
  );
}

/**
 * Resets an active process.
 */
function drunkins_job_reset_form_submit($form, &$form_state) {
  _drunkins_finish_process($form_state['values']['job_id'], array('target_state' => 'manual_reset'));
  drupal_goto('admin/config/workflow/drunkins');
}

/**
 * Form definition: display process (meta)data for a job.
 *
 * This is not really a form, there is no submit functionality - we just make it
 * one so we can use a textarea.
 */
function drunkins_view_job_metadata($form, &$form_state, $job_id) {
  if (empty($job_id)) {
    drupal_set_message(t('Cannot view process metadata for empty job!'), 'error');
  }
  $info = drunkins_job_status_info($job_id);
  if (empty($info)) {
    drupal_set_message(t('Unknown job name %process!', array('%process' => $job_id)), 'error');
  }

  // Process $info. process_settings contain everything in settings so we can
  // delete the latter, but we'll use array_diff() to be future compatible.
  if (isset($info['process_settings'])) {
    $info['settings'] = array_diff_key($info['settings'], $info['process_settings']);
    if (empty($info['settings'])) {
      unset($info['settings']);
    }
  }
  if (isset($info['process_runner_settings'])) {
    $info['runner_settings'] = array_diff_key($info['runner_settings'], $info['process_runner_settings']);
    if (empty($info['runner_settings'])) {
      unset($info['runner_settings']);
    }
  }

  // print_r is better on the eye than var_dump. However, it prints 1/''
  // for TRUE/FALSE, which can cause confusion.
  array_walk_recursive($info, function (&$value) {
    if (is_bool($value)) {
      $value = $value ? 'True' : 'False';
    }
  });
  $form['metadata'] = array(
    '#type'          => 'textarea',
    '#rows'          => 20,
    '#weight' => -99,
    '#default_value' => print_r($info, TRUE),
  );

  return $form;
}

/**
 * Form definition: display queued items for a job.
 *
 * This is not really a form, there is no submit functionality - we just make it
 * one so we can use a textarea.
 */
function drunkins_view_job_items($form, &$form_state, $job_id) {
  $items = NULL;
  if (empty($job_id)) {
    drupal_set_message(t('Cannot view process metadata for empty job.'), 'error');
  }
  $info = drunkins_job_status_info($job_id);
  if (empty($info)) {
    drupal_set_message(t('Unknown job name %process.', array('%process' => $job_id)), 'error');
  }
  else {
    /** @var DrupalReliableQueueInterface $queue */
    $queue = DrupalQueue::get($job_id);
    // $queue has no methods to peek at items...
    if (!$queue instanceof SystemQueue) {
      drupal_set_message(t('Cannot view %process items which are in a %queue_name queue.', array('%process' => $job_id, '%queue_name' => get_class($queue))), 'error');
    }
    else {
      // Max 100
      $items = db_query_range('SELECT item_id, created, expire, data FROM {queue} q WHERE name = :name ORDER BY item_id ASC', 0, 100, array(':name' => $job_id))->fetchAllAssoc('item_id');
      if (count($items) == 100) {
        drupal_set_message(t('The number of items is shown is (possibly) truncated to @number.', array('@number' => 100)), 'warning');
      }
      foreach ($items as &$item) {
        unset($item->item_id);
        $item->data = unserialize($item->data);
      }
      // print_r is better on the eye than var_dump. However, it prints 1/''
      // for TRUE/FALSE, which can cause confusion.
      array_walk_recursive($items, function (&$value) {
        if (is_bool($value)) {
          $value = $value ? 'True' : 'False';
        }
      });
    }
  }
  $form['metadata'] = array(
    '#type'          => 'textarea',
    '#rows'          => 20,
    '#weight' => -99,
    '#default_value' => print_r($items, TRUE),
  );

  return $form;
}

/**
 * Displays (or saves in Excel format) a list of items retrieved by the job.
 *
 * The 'output_format' setting is used by this function; a value of 'xls' will
 * output items in Excel format (and set HTTP headers for download.)
 * 'list_format' is commonly used by jobs' start() method as a hint for altering/
 * reformatting items for listing rather than processing; will be set to 'raw'
 * by default.
 * It is up to the individual job to (not) implement this if desired. (In other
 * words, this function does not guarantee that it will output useful data.)
 *
 * @param string $job_id
 *   ID for the job (which is equal to the queue name for queue processes),
 * @param array $settings
 *   Settings which should be passed to the job, which start() can then use for
 *   any purpose.
 *
 * @return string
 *   The output.
 *
 * @todo run start() and queue items _before_ redirecting?
 *    (Haven't thought that one through totally yet, but would solve
 *   out-of-memory issues in some cases and possibly prevent possible DDoS?
 *   Would also enable running-after-showing, if you keep the items.)
 */
function drunkins_list_items($job_id, array $settings = array()) {

  // Build settings from URL arguments, except explicitly passed.
  if (empty($settings)) {
    $settings = array_diff_key($_GET, array('q' => 1));
    if (!isset($settings['list_format'])) {
      $settings['list_format'] = 'raw';
    }
  }
  elseif (!array($settings)) {
    $settings = array('list_format' => 'raw');
  }
  // Only do drupal_set_message() for errors.
  // We'd officially need to set logging in the 'runner settings' and
  // drunkins_get_job() would then move it to settings for us. But let's
  // 'cheat' and be more direct in this case by setting this unofficial
  // logger setting, since we don't need a runner. Maybe look at this later.
  $settings['logging'] = DrunkinsJob::LOG_SCREEN;

  $job = drunkins_get_job($job_id, $settings);
  if ($job) {

    $context = array();
    $items = array();
    while (!$context || !empty($context['drunkins_process_more'])) {
      try {
        unset($context['drunkins_process_more']);
        $items = array_merge($items, $job->start($context));
      }
      catch (Exception $e) {
        // This could be any kind of exception. Its string representation could
        // give us more relevant info than just the message. However it often also
        // contains a whole backtrace (whose usefulness is limited) which will be
        // ugly if we just print it. On top of that, an exception message could be
        // unsafe, so we have to escape it, so we can't just replace newlines with
        // <br>. The next best thing is escaping the value and enclosing it in
        // <pre>.
        drupal_set_message('<pre>' . check_plain((string) $e) . '</pre>', 'error');
        drupal_goto('admin/config/workflow/drunkins/' . $job_id);
      }
      if (!is_array($items)) {
        drupal_set_message(t('Process could not start: @error', array('@error' => $items)), 'error');
        drupal_goto('admin/config/workflow/drunkins/' . $job_id);
      }

      if (!empty($context['drunkins_process_more'])) {
        // 'display_rerun' is probably a temporary solution, to be phased out
        // when we start working with queued items, as per above @todo.
        if (!empty($settings['display_rerun'])) {
          // Update context for the next call to start(). (We unfortunately
          // can't call _drunkins_run_start() because that changes run-state.)
          if (!empty($context['drunkins_rerun_start']) && is_int($context['drunkins_rerun_start'])) {
              $context['drunkins_rerun_start']++;
          } else {
              $context['drunkins_rerun_start'] = 1;
          }
          $context['drunkins_rerun_start_count'] = count($items);
          // Omit a message so that we can see where we time out, if we do.
          drupal_set_message(t('Fetched total @total items in @rerun_count iterations; trying to fetch more...',
            array('@rerun_count' => $context['drunkins_rerun_start'], '@total' => $context['drunkins_rerun_start_count'])));
        }
        else {
          drupal_set_message(t('More items need to be fetched, which is not supported here. Showing only the first batch.'), 'warning');
          if (isset($settings['cache_items']) && $settings['cache_items'] == 1) {
            drupal_set_message(t('Only a partial data set was cached. <strong>Please be careful about using the currently cached items for running a job.</strong> You may want to delete them to prevent confusion.'), 'warning');
          }
          // Break out of the loop.
          $context['drunkins_process_more'] = FALSE;
        }
      }
    }

    if (empty($items)) {
      drupal_set_message(t('There are no items to process.'));
      drupal_goto('admin/config/workflow/drunkins/' . $job_id);
    }
    else {
      $header = array();

      // Construct the table header. We need to loop through all items because
      // not all data may contain all fields.
      $prev_item = '';
      foreach ($items as $item) {
        if (is_array($item)) {
          // Item is array.
          $prev_key = '';
          foreach ($item as $key => $value) {
            if (!in_array($key, $header)) {
              if (!$prev_item || !$prev_key) {
                $header[] = $key;
              }
              else {
                // We're inserting a header for a field which was not present in
                // the previous item. Insert this header just after the previous
                // field. (This is not perfect because if two elements are
                // inserted behind the same $prev_key in different rows, they
                // may end up reordered. But it's better / the table may look
                // more logical, than if we just insert it at the end.)
                array_splice($header, array_search($prev_key, $header), 1, array($prev_key, $key));
              }
            }
            $prev_key = $key;
          }
        }
        else {
          // Item is scalar. Probably all of them are, but loop through items
          // to be sure.
          if (!in_array('Single value', $header)) {
            $header[] = 'Single value';
          }
        }
        $prev_item = $item;
      }

      if (!empty($settings['output_format'])
          && $settings['output_format'] == 'xls') {

        header("Content-type: application/vnd.ms-excel");
        header("Content-Disposition: attachment; filename=items.xls");
        header('Content-Transfer-Encoding: binary');
        header("Pragma: no-cache");
        header("Expires: 0");

        //chr(255) . chr(254) . mb_convert_encoding($table, 'UTF-16LE', 'UTF-8');
        print '<table><thead><tr><th>' . implode('</th><th>', $header) . '</th></tr></thead><tbody>';
        foreach ($items as $item) {
          $row = _drunkins_construct_list_row($item, $header);
          print '<tr><td>' . implode('</td><td>', $row) . '</td></tr>';
        }
        print '</tbody></table>';

      }
      else {
        $rows = array();
        $rows['header'] = $header;

        // Collect items.

        // Items are not check_plain()ed because they may be reformatted to
        // contain HTML. (Safe content is the job's responsibility.)

        foreach ($items as $item) {
          $rows['rows'][] = _drunkins_construct_list_row($item, $header);
        }

        $table = theme('table', $rows);

        return l('.xls download', 'admin/config/workflow/drunkins/' . $job_id . '/display',
          array('query' => array_merge($settings, array('output_format' => 'xls'))))
               . ' - '
               . l('Job settings', 'admin/config/workflow/drunkins/' . $job_id,
            array('query' => $settings))
               . $table;
      }
    }
  }
  else {
    drupal_set_message(t('Invalid job id %job', array('%job' => $job_id)), 'error');
  }
}

function _drunkins_construct_list_row($item, array $header) {
  $row = array();
  if (is_array($item)) {
    foreach ($header as $key) {
      if (!isset($item[$key])) {
        $row[] = '-';
      }
      elseif (is_array($item[$key])) {
        // We don't support printing arrays / objects.
        $row[] = '&lt;array&gt;';
      }
      elseif (is_object($item[$key])) {
        // We don't support printing arrays / objects.
        $row[] = '&lt;object&gt;';
      }
      elseif (is_bool($item[$key])) {
        $row[] = '<em>' . ($item[$key] ? 'true' : 'false') . '</em>';
      }
      else {
        // @todo make it the responsibility of a UI class to override this if it wants to print safe values.
        $row[] = check_plain($item[$key]);
      }
    }
  }
  else {
    $row['Single value'] = check_plain($item);
  }

  return $row;
}

/// Batch functionality.

/**
 * Form definition for starting the batch update process.
 */
function drunkins_batch_form($form, $form_state, $job_id) {

  // We'd officially need to set logging in the 'runner settings' and
  // drunkins_get_job() would then move it to settings for us. But let's 'cheat'
  // and be more direct in this case by setting this unofficial logger setting,
  // since we don't need a runner. Maybe look at this later. (Note we will have
  // to keep passing an argument, even if empty array, to prevent getting the
  // settings for a running process returned.)
  $job = drunkins_get_job($job_id, array('logging' => DrunkinsJob::LOG_SCREEN));
  if (!$job) {
    drupal_set_message(t('Invalid job id %job', array('%job' => $job_id)), 'error');
    return drunkins_overview();
  }
  $job_info = drunkins_job_status_info($job_id);

  $form = method_exists($job, 'settingsForm') ? $job->settingsForm() : array();

  // We allow influencing default values for known selectboxes by URL args.
  foreach (array('list_format', 'output_format', 'cache_items') as $element) {
    if (isset($_GET[$element])) {
      _drunkins_set_option_default($form, $element);
    }
  }

  $form['drunkins_job_id'] = array(
    '#type'  => 'value',
    '#value' => $job_id,
  );
  $form['cron_group'] = array(
    '#type'   => 'fieldset',
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
    '#title' => t('Delayed processing'),
    '#weight' => 100,
    '#access' => $job_info['running'] < 2,
  );
  if (!isset($job_info['cron_schedule'])) {
    $form['cron_group']['force_cron'] = array(
      '#type'   => 'checkbox',
      '#title'  => t('Force cron processing of items'),
      '#description' => t('This will have all enqueued items be processed on cron, this one time (despite cron being disabled for this job normally). If you don\'t select this option, a system outside of this module will need to process the queued items.'),
      '#default_value' => TRUE,
    );
  }
  $form['cron_group']['cron'] = array(
    '#type'   => 'submit',
    '#value'  => t('Enqueue'),
    '#suffix' => '<div class="description">This will place all items in a queue to be processed later.</div>',
  );
  $form['submit'] = array(
    '#type'   => 'submit',
    '#value'  => t('Start batch'),
    '#suffix' => '<div class="description">This will process the items immediately using Drupal\'s batch mode.</div>',
    '#weight' => 99,
    '#access' => $job_info['running'] < 2,
  );
  if ($job_info['running'] > 1) {
    $queue = DrupalQueue::get($job_id);
    drupal_set_message(t('Job %job is !status in @method method, with !n items still to be processed; it cannot be started now.',
      array(
        '%job' => $job_id,
        '!status' => $job_info['running'] == 3 ? 'running' : 'locked',
        '@method' => $job_info['method'],
        '!n'   => $queue->numberOfItems()
      )), 'warning');
    // Extra message so the string above stays consistent with one used below.
    drupal_set_message(t('See the Job Queues list for options.'), 'warning');
  }
  elseif (!empty($job_info['runner_settings']['enqueue_by_fetcher'])) {
    // Temporary (hopefully) message. @todo fix this.
    drupal_set_message('<strong>WARNING:</strong> This job (which enqueues items externally to the module) currently <strong>does not support the "Start batch" and "Display data" command</strong>. The module should be extended for this.', 'warning');
    unset($form['submit']);
  }
  return $form;
}

/**
 * Set a default value of an 'options' form element, which is somewhere in
 * $form (possibly inside a fieldset)
 *
 * @param array $form
 *  Form definition
 * @param $element_name
 *  Element name. $_GET[$element_name] must hold the value to set.
 */
function _drunkins_set_option_default(&$form, $element_name) {
  if (isset($form[$element_name])) {

    // Found. Check whether this is a valid default value; if so, set.
    if (isset($form[$element_name]['#options'][$_GET[$element_name]])) {
      $form[$element_name]['#default_value'] = $_GET[$element_name];
    }
  }
  else {

    // Search in fieldsets.
    foreach ($form as &$element) {
      if (isset($element['#type']) && $element['#type'] == 'fieldset') {
        _drunkins_set_option_default($element, $element_name);
      }
    }
  }
}

/**
 * Submit callback.
 * Starts the batch update process.
 *
 * Note: various submit buttons may redirect here; some of them are defined
 * inside classes' settingsForm().
 */
function drunkins_batch_form_submit(&$form, &$form_state) {

  $job_id = $form_state['values']['drunkins_job_id'];
  // Usually we want to see messages on screen if we submit from a form. We
  // specify this unless the job config explicitly set it in 'runner_settings'.
  $info = drunkins_job_status_info($job_id);
  $logging = isset($info['runner_settings']['logging']) ? $info['runner_settings']['logging'] : DrunkinsJob::LOG_SCREEN | DrunkinsJob::LOG_WATCHDOG;

  // Get 'process settings' from all form values; clean out system values.
  $settings = array_diff_key($form_state['values'], array_flip(array(
    'form_build_id',
    'form_id',
    'form_token',
    'op',
    'submit',
    // ...plus values from submit buttons (which aren't settings)
    'cron',
    'display',
    // ...plus settings we know we don't need in $settings.
    'drunkins_job_id',
    'force_cron',
  )));
  // @todo remove 'logging' from settings, and have it only in runner_settings,
  //   when log() gets rewritten or removed from the job class. Also remove it
  //   from the job settingsForm() if possible; it is only there because we
  //   don't have another easy way to
  $settings['logging'] = $logging;
  // We assume most 'runner settings' are not in this form (unless hardcoded
  // below), so start off with an almost empty array. (Any settings defined in
  // queue info hooks will be merged in drunkins_start_job().)
  $runner_settings = array(
    'logging' => $logging
  );
  $reset_logger_settings = NULL;

  switch ($form_state['input']['op']) {

    case t('Enqueue'):
      if (!empty($form_state['values']['force_cron'])) {
        $runner_settings['force_cron'] = TRUE;
      }
      $reset_logger_settings = DrunkinsJob::LOG_WATCHDOG;
      // no break!
    case t('Start batch'):
      // We always want to see "There are no items to process." so override any
      // log severity for that message.
      $runner_settings['log_severity_no_items'] = isset($runner_settings['log_level']) ? $runner_settings['log_level'] : WATCHDOG_INFO;
      // 'list_format' could influence the output of start(), so clear too.
      unset ($settings['list_format']);

      drunkins_start_job($job_id,
        $form_state['input']['op'] == t('Start batch') ? 'batch' : 'queue',
        $runner_settings, $settings);

      if ($reset_logger_settings) {
        // For the remainder of the process, logging should be different.
        // The process state, including runner settings, have not been saved
        // into the database yet, so assume that will get done automatically.
        $settings['logging'] = $reset_logger_settings;
        _drunkins_set_process_config($job_id, 'job', $settings);
        $runner_settings['logging'] = $reset_logger_settings;
        _drunkins_set_process_config($job_id, 'runner', $runner_settings);
      }
      break;

    default:
      // We'd officially need to set logging in the 'runner settings' and
      // drunkins_get_job() would then move it to settings for us. But let's
      // 'cheat' and be more direct in this case by setting this unofficial
      // logger setting, since we don't need a runner. Maybe look at this later.
      $settings['logging'] = DrunkinsJob::LOG_SCREEN | DrunkinsJob::LOG_WATCHDOG;
      $job = drunkins_get_job($job_id, $settings);
      if (method_exists($job, 'settingsFormSubmit')) {
        $output = $job->settingsFormSubmit($form_state['input']['op']);

        // If the method returns a string, this is assumed to be HTML output.
        // We are however in the middle of a form submit function -- so we
        // will add this output to the form, and not redirect. The usual form
        // will be rendered below our output.
        if (is_string($output)) {
          $form['output'] = array(
            '#type'   => 'markup',
            '#markup' => '<h2>'
                         . t('The %a action produced the following output:',
                             array('%a' => $form_state['input']['op']))
                         . '</h2><hr/>' . $output . '<div class="description">&lt' . t('end output') . '&gt;</div><hr/>',
            '#weight' => -99,
          );
          $form_state['no_redirect'] = TRUE;
        }
      }
      else {
        drupal_set_message(t('Unknown operation %op!', array('%op' => $form_state['input']['op'])), 'error');
      }
  }
}

/**
 * queues operations into a batch set.
 */
function _drunkins_batch_enqueue_items($job_id, array $items, array $job_context, $non_progressive = FALSE, $title = '', $return_path = '') {
  /* 'drunkins_rerun' (int) means we're starting a new batch for the Nth time,
   * after finish() has run. In that case, 'drunkins_process_more' is not set
   * and 'drunkins_rerun_start' has no meaning (additional calls to start() are
   * not supported at this point).
   * If 'drunkins_rerun' not set:
   * - 'drunkins_process_more' means another start() call should be made next.
   * - 'drunkins_rerun_start' (int) indicates how many calls have already been
   *    made, unless we're making only one call; then it's unset.
   * - 'drunkins_rerun_start_count' (int) indicates the number of items
   *    already fetched in previous runs.
   *
   * If drunkins_rerun_start > 1, we will not add a second/third set to the
   * existing batch, but tag the operations onto the same set instead, because:
   * - otherwise we would have to take special measures to carry 'our context'
   *   over to the next batch set (which for us is an arbitrary split of a bunch
   *   of drunkins_batch_process() calls);
   * - working with batch sets isn't logical for us anyway. If we add extra sets
   *   from the drunkins_batch_rerun_start() call (i.e. while running the
   *   first set), the batch_set() will always add an additional set with index
   *   1, and queue the extra items with index 1 - meaning extra batch_set()
   *   calls will add sets in reverse, and all those sets except the last one
   *   will be useless, without items.
   *   (In other words, I think batch_set() is essentially buggy / the
   *   array_slice commands in there are useless, because the queued items are
   *   not moved together with the sets.)
   *
   * (In other words: this function abstracts away the way of queueing
   * operations, but in practice this means that we will execute separate code
   * based on whether we are called from drunkins_batch_rerun_start().)
   */

  $create_new_batch = empty($job_context['drunkins_rerun_start'])
               || $job_context['drunkins_rerun_start'] == 1
               || !empty($job_context['drunkins_rerun']);
  if ($create_new_batch) {

    // Set up the set with operations. Lose any non-numeric keys for the items.
    // (We could also choose to have $items repeated operations, all calls to
    // drunkins_batch_process() with one item. That will use more memory in
    // this array while queueing but less memory while dequeueing, and it will
    // guard against $items being too big for the queue.data field. But we'll
    // have to do some reorganizing in a.o. drunkins_batch_process() for that?
    // (Discontinue the use of sandbox anywhere?)
    $batch_set = array(
      'operations' => $items ? array(
        array('drunkins_batch_process', array($job_id, array_values($items))),
      ) : array(),
      'file' => drupal_get_path('module', 'drunkins') . '/drunkins.run.inc',
      'finished' => 'drunkins_batch_finished',
      // We can't set sandbox & results here, because these are overwritten by
      // batch_set(). But we can populate other (undocumented) array elements;
      // these will be stored by _batch_populate_queue() and passed to the
      // worker callbacks.
      'job_context' => &$job_context,
    );

    if (empty($title)) {
      $info = drunkins_job_status_info($job_id);
      if (!empty($info['title'])) {
        $batch_set['title'] = $info['title'];
      }
    }
    else {
      $batch_set['title'] = $title;
    }
    if (!empty($job_context['drunkins_rerun'])) {
      $batch_set['init_message'] = t('Re-starting process (@rerun_count) with @count items.',
        array('@rerun_count' => $job_context['drunkins_rerun'], '@count' => count($items)));
    }
    elseif (!empty($job_context['drunkins_rerun_start'])) {
      $batch_set['init_message'] = t('Fetched @count items; trying to fetch more...',
        array('@count' => count($items)));
      // Add function to rerun start() in the same batch set.
      array_unshift($batch_set['operations'],
        array('drunkins_batch_rerun_start', array($job_id)));
    }
    else {
      // The standard progress message is "processing 0 of 1" (because we have 1
      // batch set/step, containing many items. The batch message 'M ::: N' is
      // displayed just below it). Let's make it a bit nicer.
      $batch_set['progress_message'] = t('Processing items...');
    }

    batch_set($batch_set);
    if ($non_progressive) {
      // The definition is set as a 'set' _inside_ a batch. We need to get the
      // whole batch, and set 'progressive' there (in outer definition array).
      $batch = &batch_get();
      $batch['progressive'] = FALSE;
    }
    if ($return_path) {
      // Same for return path. (Won't have effect for non-progressive batches.)
      $batch = &batch_get();
      $batch['source_url'] = $return_path;
    }
  }
  else {
    // We only end up here while executing drunkins_batch_rerun_start().
    $batch = &batch_get();
    if (empty($batch['id']) || count($batch['sets']) != 1
        || !isset($batch['sets'][0]) || isset($batch['sets'][0]['operations'])) {
      // @todo when runner_settings are stored for batch processes too during the
      //   lifetime of a process, this line must be deleted.
      _drunkins_set_process_config($job_id, '*runner_batch_logging_hack', array());
      _drunkins_log($job_id, t('Internal code error while adding extra items to batch; discarding items!'), array(), WATCHDOG_CRITICAL);
    }
    else {
      // Stick new operation into the existing set and queue it with the same
      // index as the existing ones. It will be executed after the ones already
      // queued (but before the finish callback).
      if ($items) {
        $batch['sets'][0]['operations'] = array(
          array('drunkins_batch_process', array($job_id, array_values($items)))
        );
        _batch_populate_queue($batch, 0);
        // Above is one extra queued operation (there are now multiple
        // 'drunkins_batch_process' operations); account for this.
        $batch['sets'][0]['total']++;
        $batch['sets'][0]['count']++;
        // $job_context is an alias into $batch['sets'][0]['job_context']
      }
    }
    if (!empty($job_context['drunkins_process_more'])) {
      $batch['sets'][0]['progress_message'] = t('Fetched @count items (total @total in @rerun_count iterations); trying to fetch more...',
        array('@rerun_count' => $job_context['drunkins_rerun_start'], '@count' => count($items), '@total' => $job_context['drunkins_rerun_start_count']));
    }
    else {
      // progress_message is not set unless drunkins_process_more was true at
      // one point. However, if it's set, the batch API propagates it to the
      // next step (processing). So overwrite it.
      $batch['sets'][0]['progress_message'] = t('Processing items...');
    }
  }
}

/**
 * 'worker callback' for batch process: get more items.
 */
function drunkins_batch_rerun_start($job_id, array &$batch_context) {

  // First time: set up the 'sandbox' and 'results' (persistent data between
  // calls), which we've stored in the batch set's data.
  if (!_drunkins_batch_ensure_context($batch_context)) {
    // This is almost impossible. Abort without specification.
    return;
  }

  // @todo when runner_settings are stored for batch processes too during the
  // lifetime of a process, this line must be deleted.
  _drunkins_set_process_config($job_id, '*runner_batch_logging_hack', array());

  $runner_settings = _drunkins_get_process_config($job_id);

  try {
    $items = _drunkins_run_start($job_id, $batch_context['results']);
  }
  catch (\Exception $e) {
    // The process failed; is already canceled. The queue is cleaned up.
    //@todo test this. I don't believe it's true because the queue did not
    //  contain items in the first place. (Batch API does its own queueing, and
    //  we just let the items sit in the batch queue unreachable, for now.
    //  Something will clean them up.)
    //  According to the comment just below, $items will be empty -and I have
    //  apparently tested that once. But I don't see anymore, how. So test this
    //  again and redo the comment so I will remember. OR debug this.
    _drunkins_log($job_id, 'Rerunning start for job "@job" threw exception; deleting all previously gathered items from the job. Exception: @error', array('@job' => $job_id, '@error' => $e->getMessage()), WATCHDOG_ERROR);
    // drunkins_batch_process() will still be called as part of the running
    // batch, but it can handle empty $items. drunkins_batch_finished() will
    // also be called and we should make sure it does not call the job's
    // finish() method. Smuggle a flag in the one value that we know is never
    // set here on normal operation:
    $batch_context['results']['drunkins_process_more'] = '*error_on_restart';
    return;
  }

  if (!empty($batch_context['results']['drunkins_process_more'])) {
    $batch_context['finished'] = 0;
  }
  if (empty($runner_settings['enqueue_by_fetcher'])) {
    // @todo once we actually implemented 'enqueue_by_fetcher' for batch processes, check whether this if() logic is OK.
    _drunkins_batch_enqueue_items($job_id, $items, $batch_context['results']);
  }
}

/**
 * Ensures that our results/sandbox values are filled.
 */
function _drunkins_batch_ensure_context(&$batch_context) {

  // 'results' is empty every time we start a new set. We've made sure our
  // 'job_context' always has one value set, so we can recognize if it's filled.
  if (empty($batch_context['results'])) {
    $current_set = &_batch_current_set();
    if (!isset($current_set['job_context'])) {
      // fatal.
      return FALSE;
    }
    $batch_context['results'] = $current_set['job_context'];
  }

  // 'sandbox' is empty every time we call a new operation.
  if (empty($batch_context['sandbox'])) {
    // One call to this function will process a fixed number of rows:
    // @todo check / document whether this should be a setting - bigger number makes for less calls by Core JS 'batch runner code'?
    if (!isset($batch_context['sandbox']['limit'])
        || !is_numeric($batch_context['sandbox']['limit'])) {
      $batch_context['sandbox']['limit'] = 5;
    }
    $batch_context['sandbox']['progress'] = 0;
  }
  return TRUE;
}

/**
 * 'worker callback' (several items) for batch process: process items.
 */
function drunkins_batch_process($job_id, array $items, array &$batch_context) {
  if (!$items) {
    // Exit immediately and do not return here.
    return;
  }

  $job = drunkins_get_job($job_id);
  if (!isset($job)) {
    // Should we throw an error here? Which? This is very unlikely to happen.
    return;
  }
  // First time: set up the 'sandbox' and 'results' (persistent data between
  // calls), which we've stored in the batch set's data.
  if (!_drunkins_batch_ensure_context($batch_context)) {
    // This is almost impossible. Abort without specification.
    return;
  }

  // @todo actually make this work by persisting runner settings / test this.
  $runner_settings = _drunkins_get_process_config($job_id);
  if (!empty($runner_settings['memory_limit'])) {
    _drunkins_set_memory_limit($job_id, $runner_settings['memory_limit']);
  }

  $count = count($items);
  if (empty($batch_context['sandbox']['limit']) || !is_numeric($batch_context['sandbox']['limit'])
      || $batch_context['sandbox']['limit'] <= 0) {
    $batch_context['sandbox']['limit'] = $count;
  }
  $limit = min($batch_context['sandbox']['progress'] + $batch_context['sandbox']['limit'], $count);
  $i = $batch_context['sandbox']['progress'];
  while ($i < $limit) {
    $batch_context['results']['drunkins_progress'] = ++$i;
    try {
      $job->processItem($items[$i - 1], $batch_context['results']);
    }
    catch (\Exception $e) {
      if (isset($runner_settings['handle_process_exception'])
          && $runner_settings['handle_process_exception'] === 'discard') {
        // Do nothing except accounting. The item will be dequeued.
        _drunkins_log($job_id, 'Processing "@job" item failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
        $batch_context['results']['drunkins_exception_count']++;
      }
      else {
        // Default value: 'fallthrough'. (Other values should have been caught
        // before process start.) For batch, this usually means the caller will
        // display the exception as an error message, in a screen with a
        // "Please continue to the error page." link. (The target page will
        // run the 'finished' callback which will not run finish() but reset the
        // process state - and display another message about an exception).
        throw $e;
      }
    }
  }
  $batch_context['sandbox']['progress'] = $i;

  // Set status message and progress indicator, whether or not we've finished.
  // This function will be called again when $batch_context['finished']
  // is explicitly set to a value < 1. (It's initalized to 1 by the caller.)
  if ($count == 0) {
    $batch_context['message'] = 'Aborted.';
  }
  else {
    $batch_context['finished'] = $i / $count;
    $batch_context['message'] = $i . ' ::: ' . $count;
  }

  // This one is necessary for the 'finished callback'; set it every time to
  // protect against messing by processItem().
  $batch_context['results']['drunkins_name'] = $job_id;
}

/**
 * 'finished callback' for batch process.
 */
function drunkins_batch_finished($success, array $job_context, $operations) {
  if (!empty($job_context['drunkins_name'])) {
    $job_id = $job_context['drunkins_name'];
  }
  // Sometimes when a fatal error has happened, $job_context (in other words,
  // batch_set['results']) is empty. Check if $operations still contains the
  // 'drunkins_batch_process' call with items. If so, we can get the job ID from
  // there.
  elseif (isset($operations[0][0]) && $operations[0][0] == 'drunkins_batch_process'
          && isset($operations[0][1][0])) {
    $job_id = $operations[0][1][0];
  }

  if (isset($job_id)) {
    // Set default suitable for logging.
    // @todo We likely won't have any 'process_runner_settings' at the moment
    //   so we'll always fall back to the default. When '*runner_settings' get
    //   remembered for batch jobs, this default can go away.
    _drunkins_set_process_config($job_id, '*runner_batch_logging_hack', array());
  }

  $message = '';
  $non_progressive_message = '';

  // The 'success' parameter means no fatal PHP errors were detected. All
  // other error management should be handled using $job_context.
  if ($success) {
    if (isset($job_context['drunkins_process_more'])
        && $job_context['drunkins_process_more'] === '*error_on_restart') {
      // _drunkins_run_start actually logged a message so this is duplicate.
      // Hack in a message for the non-progressive case which demands a return
      // value at the moment.
      // @todo reevaluate after rewrite of return values?
      $non_progressive_message = t('Starting job "@job" was done in multiple phases; one of these phases failed.', array('@job' => $job_id));
    }
  }
  if (!$non_progressive_message) {
    if (!isset($job_id)) {
      _drunkins_log('?', "Batch 'finished callback' could not derive job id, so could not properly finish the job.", array(), WATCHDOG_ERROR);
      $job_id = '?';
    }
    if (!$success) {
      _drunkins_log($job_id, 'Job "@job" encountered a fatal error and did not properly finish.', array('@job' => $job_id), WATCHDOG_ERROR);
    }
    else {
      // Clear out the 'continue' value, to make sure we'll only use it when
      // finish() explicitly sets it. (Theoretically a processItem() could set
      // it, though we don't support this yet; we'll protect against it and
      // won't warn.)
  // @todo consider if we want to make this array (part of) the return
  // argument, instead of the message (if we want to get rid of that anyway)
      unset($job_context['drunkins_process_more']);
      $job = drunkins_get_job($job_id);
      if (!$job) {
        _drunkins_log($job_id, "Batch 'finished callback' failed to create job \"@job\"", array('@job' => $job_id), WATCHDOG_ERROR);
      }
      else {
        // Log before finish() (at debug level).
        // @todo what do we exactly want to do immediately after, when finish() does not return a message anymore?
        _drunkins_log($job_id, 'Calling finish() for job "@job".', array('@job' => $job_id), WATCHDOG_DEBUG);
        $runner_settings = _drunkins_get_process_config('runner');
        if (!empty($runner_settings['memory_limit_finish']) || !empty($runner_settings['memory_limit'])) {
          _drunkins_set_memory_limit($job_id, !empty($runner_settings['memory_limit_finish'])
            ? $runner_settings['memory_limit_finish'] : $runner_settings['memory_limit']);
        }
        try {
          $message = $job->finish($job_context);
        }
        catch (\Exception $e) {
          //@TODO if we make the items the return value, we can simplify a lot below because we cannot generate an exception AND more items.
          //  We'll also have to think about the other @todo below though.
          if (isset($runner_settings['handle_finish_exception'])
              && ($runner_settings['handle_finish_exception'] === 'continue'
                  || $runner_settings['handle_finish_exception'] === 'finish')) {
            // Log exception and finish normally (or continue on more items).
            _drunkins_log($job_id, 'Calling finish() for job "@job" failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
            if (!empty($job_context['drunkins_process_more']) && is_array($job_context['drunkins_process_more'])) {
              if ($runner_settings['continue_on_finish_exception'] === 'continue') {
                _drunkins_log($job_id, '"@job" finish() generated @count more items to process: continuing despite exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
              }
              else {
                _drunkins_log($job_id, '"@job" finish() generated @count more items to process: discarding them because of exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
                unset($job_context['drunkins_process_more']);
              }
            }
          }
          else {
            // Default value: 'break'. (Other values should have been caught
            // before process start.) We now have two options:
            // - continue without running _drunkins_finish_process() (The
            //   standard UI will redirect to the start screen again, where the
            //   job cannot be started again; the current process will be stuck
            //   with 0 items to process, until a user manually resets it.
            // - throw the exception, which will have Drupal generate a plain
            //   "unexpected error; please try again" screen without any links
            //   to e.g. finish the batch process. (The URL still includes
            //   '&op=finished' so releasing the page will indeed run finish()
            //   again. If the exception persists, the user must explicitly
            //   navigate away from the page and the current process is still
            //   stuck like it would be for the first option.)
            // We choose the latter because it's more 'in your face'.
            // (Note if we ever get an option to run finish() for batch
            // processes with 0 items like we do for queue processes... this
            // whole thing needs to be reevaluated. See also
            // _drunkins_queue_process_check_finish() for that case.)

            // If a process returned more items to process, we'll always need to
            // log that fact here, despite not always logging the exception itself
            // here (if the caller takes care of it). This means the "discarding
            // items" log will happen a bit out of order.


            // We know (or assume) the caller will log the exception. We don't
            // know what function it will log as the source though, so log the
            // fact that the exception was thrown in finish().
            _drunkins_log($job_id, 'Calling finish() for job "@job" generated an exception.', array('@job' => $job_id), WATCHDOG_NOTICE);
            // We're not logging about any extra items; they *probably* will be
            // saved into context (if the PHP 'shutdown functions' still run
            // after an exception; haven't tested that) and be processed if the
            // user reloads and triggers finish() and that finishes without
            // throwing an exception next time.
            // @todo this completely changes if we make the extra items a return
            //   value instead of a subkey in $context. Think about this.

            throw $e;
          }
        }
      }
    }
  }

  $non_progressive = &drupal_static('drunkins_run_non_progressive');

  // From here on, job ID will only be used for logging.
  if (!isset($job_id)) {
    $job_id = '?';
  }
  if (!empty($job_context['drunkins_process_more']) && is_array($job_context['drunkins_process_more'])) {
    // If we got any message back, don't just drop it; log/print it. (Print
    // by default, unlike everywhere else.)
    if ($message) {
      _drunkins_log($job_id, $message, array(), WATCHDOG_INFO);
    }
// @todo maybe make the 'executive decision' that status logs like the below
// will not be done by the runner, and logging is the exclusive responsibility
// of the job (except in clear errors)? Not sure; we should treat it the same as the 'N items received' from start().
    _drunkins_log($job_id,
      'finish() for job "@job" returned @count new items; continuing...',
      array('@job' => $job_id, '@count' => count($job_context['drunkins_process_more'])),
      WATCHDOG_INFO);

    $batch = &batch_get();
    // We're being called from _batch_finished(). That function processes things
    // differently depending on whether the batch is progressive:
    if (empty($non_progressive)) {
      // Remember context because our caller will clear out the batch variables.
      $continue_context = &drupal_static('drunkins_batch_continue', array());
      $continue_context = $job_context;

      // Intercept the caller by making it pass through drupal_redirect_form()
      // (which is always called even though we are on a regular page, not a
      // form) and to our own callback.
      $batch['form_state']['no_redirect'] = TRUE;
      $batch['redirect_callback'] = 'drunkins_batch_continue';
      // We need to remember the source URL to redirect to after continuing, and
      // $batch['url'] does not have any function anymore now that we don't
      // redirect automatically. So remember it in $batch['url']; then it will
      // get passed to the redirect callback.
      $batch['url'] = $batch['source_url'];
    }
    else {
      // _batch_finished() will just exit for non-progressive batches, so we
      // don't need to do funny overriding, and have to continue by ourselves.
      drunkins_batch_continue($batch['source_url'], $job_context);
    }
  }
  else {
    if (!empty($non_progressive)) {
      // Reset memory limit to initial one. Do it here so it's done only once if
      // we were called recursively through drunkins_batch_continue(). For
      // progressive batches we don't care because the caller will redirect.
      _drunkins_set_memory_limit($job_id);

      // Since no batch operations/finish callback return anything to the
      // caller, pass the context back through a static variable.
      $log = &drupal_static('drunkins_last_log');
      $log = array($non_progressive_message ? $non_progressive_message : $message, array(), WATCHDOG_INFO, $job_context);

      $non_progressive = FALSE;
    }
    // For 'normal' jobs, report message here. If the process did not return a
    // message, that's fine.
    elseif (!empty($message)) {
      drupal_set_message($message);
    }

    // In the '*error_on_restart' case we already canceled the process. (We just
    // get here because we couldn't delete process/finish callbacks anymore.)
    if (!isset($job_context['drunkins_process_more'])
        || $job_context['drunkins_process_more'] !== '*error_on_restart') {
      _drunkins_finish_process($job_id, array('delete_items' => FALSE));
    }
  }
}

/**
 * Start up a new batch process with items received from a previous batch'
 * 'finished' operation.
 */
function drunkins_batch_continue($path, array $job_context) {
  $non_progressive = drupal_static('drunkins_run_non_progressive');

  if (!$non_progressive) {
    // We're called as 'redirect_callback', so we need to get our argument from
    // a stored static variable.
    $job_context = drupal_static(__FUNCTION__);

    // The difference between this function and drunkins_batch_start() is, we
    // are not called from a submit function so the grandparent process_form()
    // does not call calls batch_process().
    // Now that we overrode the 'drupal_goto' to get here, our caller will do
    // nothing, return nothing, output no HTML. So we will call batch_process()
    // ourselves, which initializes a new batch with the above set and calls
    // drupal_goto().
    // The batch has been removed from the database table. We don't need it; we
    // can just call batch_set() for a new batch, and batch_process() will fill
    // the database table and start things as a new process with incremented ID.
  }
  else {
    // For a progressive batch, _batch_finished() has deleted the finished batch
    // and calling batch_set() will create a new batch. For us, it still exists;
    // we need to delete it. (We can't add another set to the existing batch,
    // because _batch_finished() would end up calling the 'finished callback'
    // for the first set again.)
    $batch = &batch_get();
    $batch = NULL;

    // Since we don't do funny overrides or redirects, this creates a recursive
    // call tree: batch_process() -> _batch_process() -> _batch_finished() ->
    // drunkins_batch_finished() -> drunkins_batch_continue() ->
    // batch_process(). All these functions will return immediately, after
    // drunkins_batch_finished() yields no items to process. Except
    // _batch_finished(), which tries to do some $_SESSION cleanup. We could try
    // to do proper $_SESSION cleanup because we're changing the batch ID before
    // executing that code... but it's buggy anyway, so it's clearly not used.
  }

  // Set a new batch.
  // Unset drunkins_process_more because we don't want it stored.
  $items = $job_context['drunkins_process_more'];
  unset($job_context['drunkins_process_more']);
  if (isset($job_context['drunkins_rerun']) && is_numeric($job_context['drunkins_rerun'])) {
    $job_context['drunkins_rerun']++;
  }
  else {
    $job_context['drunkins_rerun'] = 1;
  }
  _drunkins_batch_enqueue_items($job_context['drunkins_name'], $items, $job_context, $non_progressive,
    !empty($job_context['drunkins_process_more_title']) ? $job_context['drunkins_process_more_title'] : '',
    $path
  );
  unset($items);

  if ($non_progressive) {
    // Reset memory limit from the 'finish' one, before doing more processing
    // but after queueing the items. For progressive batches we're not in the
    // same HTTP request anymore as when calling finish().
    // @todo actually make this work by persisting runner settings / test this.
    $runner_settings = _drunkins_get_process_config($job_context['drunkins_name']);
    if (!empty($runner_settings['memory_limit_finish'])) {
      if (!empty($runner_settings['memory_limit'])) {
        $initial_limit = _drunkins_get_initial_memory_limit();
        _drunkins_set_memory_limit($job_context['drunkins_name'], $initial_limit > $runner_settings['memory_limit'] ? $initial_limit : $runner_settings['memory_limit'], TRUE);
      }
      else {
        _drunkins_set_memory_limit($job_context['drunkins_name']);
      }
  }
  }

  batch_process();
}

/**
 * Claims a job process, runs the job's start() method and initializes all
 * related job info (context, run state, database table, etc). This function
 * abstracts away the icky details around keeping the cached job info and
 * database table in sync while the job process' state changes.
 *
 * It's meant to only be called from drunkins_start_job(); it's only separated
 * to isolate these icky details and because drunkins_start_job() gets too long.
 *
 * @param string $job_id
 * @param string $method
 * @param array $context
 *   Initial context (usually an empty array) passed by reference if the caller
 *   needs it. (The reference is usually for batch processes because this
 *   function stores the context for queue processes but not for batch.)
 *
 * @return array
 *   Items as returned by the job's start() method.
 *
 * @throws \Exception
 *   Any exception or own error condition encountered. This function will
 *   already log the error so the caller will not have to do much with it (which
 *   is good because this function is an integrated part of the caller).
 *
 * @see drunkins_start_job()
 */
function _drunkins_start_process($job_id, $method, array &$context) {

  // Get $settings and $runner_settings into variables just because we need to
  // store them twice in a related array. (If we start storing them in separate
  // database fields, the code may change.)
  $runner_settings = _drunkins_get_process_config($job_id);
  $settings = _drunkins_get_process_config($job_id, 'job');

  if ($method === 'batch') {
    $init_data = array('settings' => $settings);
    // For batch runs, we don't store 'context' in the drunkins_process table;
    // it will be stored somewhere else (likely the batch table, or a
    // MemoryQueue for progressive batches through batch_set()) by the caller.
  }
  else {
    // drunkins_progress for now is a kind-of-half-officially-documented
    // progress indicator; it's 1-based (1 for the first processed item). For
    // batch processes it just copies the 'sandbox][progress' value but for
    // queue processes we need to initialize it here.
    // We may revisit and/or document this better, once we figure out what to
    // do with concurrency of processing items (which is not supported - but we
    // mostly ignore it until now and will properly document things later).
    // @todo make docs about concurrency of processing items etc
    $context['drunkins_progress'] = 0;
    $init_data = array(
      'settings' => $settings,
      'context' => $context,
      'runner_settings' => $runner_settings
    );
  }
  $init_fields = array(
    'method' => $method,
    'run_info' => 1,
    'started' => time(),
  );
  $calc_init_fields = array(
    'running' => 3,
  );

  // Gather fields for DB update.
  $db_fields = $init_fields + array('data' => serialize($init_data));
  // Try to update the run status of the process in a 'safe' way: only one
  // thread can succeed in UPDATEing the same row.
  $status = db_update('drunkins_process')
    ->fields($db_fields)
    ->condition('queue_name', $job_id)
    ->where('((run_info = 0 AND finished >= started) OR run_info > 3)')
    ->execute();
  if (!$status) {
    // We didn't succeed in updating a row; maybe there is none for $job_id yet.
    // Try to insert one. If another PHP process just claimed the job, this
    // will throw a database exception. Since our table has no serial ID, the
    // return value of the insert query is always 0 - so we will need to check &
    // lock the database row with a repeated update query.
    db_insert('drunkins_process')
      ->fields(array(
        'queue_name' => $job_id,
        'method'     => $method,
        'run_info'   => 0,
        'started'    => 0,
        'finished'   => 0,
        'data'       => serialize(array()),
      ))
      ->execute();
    $status = db_update('drunkins_process')
      ->fields($db_fields)
      ->condition('queue_name', $job_id)
      ->where('((run_info = 0 AND finished >= started) OR run_info > 3)')
      ->execute();
  }
  if (!$status) {
    throw new \Exception('Failed to claim job; drunkins_process row update failed somehow. This is strange.');
  }

  // Update the cached job status info to match the database state.
  // (Otherwise in the remainder of this HTTP request, drunkins_get_job()
  // will get the wrong settings, drunkins_cron_process_queue_item() will
  // get no context and discard items, etc.) This is not very important (we
  // probably won't use the data) until the job actually starts running, but we
  // try to be as complete/early as possible to prevent obscure bugs. (Also,
  // now that we do that anyway, we'll use the info in _drunkins_run_start().)
  $info = &drunkins_job_status_info($job_id);
  // Set 'running' already; necessary to make both _drunkins_run_start() and
  // _drunkins_finish_process() behave correctly.
  $info = array_merge($info, $init_fields, $calc_init_fields, array(
    'process_settings' => $settings,
    // Set context by reference so $context is modified after exiting this
    // function, but use $info['process_context'] below. To be changed when
    // this function is retired.
    'process_context' => &$context,
    'process_runner_settings' => $runner_settings,
  ));

  // Unset these context values on first run, to prevent strange conditions.
  // Jobs should not change these.
  unset($info['process_context']['drunkins_rerun_start']);
  unset($info['process_context']['drunkins_rerun_start_count']);

  $items = _drunkins_run_start($job_id, $info['process_context']);

  // A second kind-of-half-officially-documented indicator for processing runs:
  $info['process_context']['drunkins_exception_count'] = 0;

  $items_queued = empty($runner_settings['enqueue_by_fetcher']) ? count($items) : DrupalQueue::get($job_id)->numberOfItems();
  if (!$items_queued && empty($info['process_context']['drunkins_process_more'])) {
    _drunkins_log($job_id, 'There are no items to process for job "@job".', array('@job' => $job_id),
      isset($runner_settings['log_severity_no_items']) ? $runner_settings['log_severity_no_items'] : WATCHDOG_INFO);
    _drunkins_finish_process($job_id, array('target_state' => 'empty_start', 'delete_items' => FALSE));
  }
  elseif ($method !== 'batch') {
    // Also if start() reset $context, we still want to set 'progress'.
    $info['process_context']['drunkins_progress'] = 0;
    if ($info['process_context'] !== $init_data['context']) {
      // Context has been changed by start(), so it must be updated in the
      // database. We'll postpone this if possible, because the context may
      // still be used & changed by drunkins_cron_process_queue_item(). Call the
      // 'current context' function with this job ID to mark its context as
      // needing to be saved. (This registers a shutdown function. If we started
      // another job in the same HTTP request, that job's context will be
      // written to the database now. So if we initialize job A and job B, and
      // then process queued items for job A and B, in one HTTP request... we
      // make two extra database calls for writing context, which could be
      // optimized away. But that's not worth the effort.)
      drunkins_current_context($job_id);
    }
  }

  return $items;
}

/**
 * Runs start() for a jobs; handles some errors.
 *
 * This also sets drunkins_process_more/_rerun_start values in $job_context.
 * Normally, no logging is done about number of items returned/need to continue;
 * that's up to the caller (and should be done when items are actually queued).
 * This function only logs before calling start (severity: debug), logs that
 * items are queued if 'enqueue_by_fetcher' is set, and logs (most, not all)
 * 'fatal' errors. In error cases, the process info (in context and db) is reset
 * if this is NOT the first start() run; if it is the first, the caller is
 * responsible for that.
 * @todo revisit above once we make this into a runner class? At the moment, the
 *  full info for saying 'status is batch-error-on-start vs queue-error-on-start
 *  is not known to us. Hence the inconsistency in process info resetting.
 *
 * Besides the input arguments, this function also uses (a subset of)
 * process_runner_settings in status info.
 *
 * This function changes 'drunkins_rerun_start(_count)' values in $job_context.
 * The 'drunkins_process_more' value which can be set by the job class, is used
 * as input for this. (Also, it always unsets 'drunkins_override_fetch'.)
 *
 * @param $job_id
 *   Job ID / machine name for the the queue.
 * @param array $job_context
 *   Context, which will be altered by the job and this function. (Note that
 *   even though other things will be altered directly into the job info, we
 *   must use this argument, not the job info, to alter context.)
 *
 * @return array
 *   Items as returned by the job's start() method.
 *
 * @throws \Exception
 *   Any exception or own error condition encountered. This function will
 *   already log the error. Process info (in context and db) is reset if this is
 *   NOT the first start() run; if it is, the caller is responsible for that.
 * @todo revisit above once we make this into a runner class? At the moment, the
 *  full info for saying 'status is batch-error-on-start vs queue-error-on-start
 *  is not known to us. Hence the inconsistency in process info resetting.
 */
function _drunkins_run_start($job_id, array &$job_context) {

  // Remember if this is the very first run in the process (which means:
  // - the process never indicated 'process_more' - from start(), processItem()
  //   or finish();
  // - the caller is _drunkins_start_process().
  // (If these 2 things are not equivalent, this comment should be rewritten :))
  $first_run = empty($job_context['drunkins_rerun_start']);
  // Note the job class itself is not meant to set 'drunkins_rerun_start' and
  // '..._count' values but it can use them; it means how many times start() had
  // already been called / how many items had been returned in _previous_ calls.
  // If no calls have been made, both are unset.

  $job = drunkins_get_job($job_id);
  if (!isset($job)) {
    // This will happen e.g. if the 'class' was not defined.
    throw new \Exception( 'Failed to create job class instance.');
  }

  // If we have custom items, pass them to start() in context.
  $items = _drunkins_get_process_config($job_id, 'items');
  if ($items !== FALSE) {
    $job_context['drunkins_override_fetch'] = $items;
    // We don't know how many items these can be, so for memory management,
    // unset these.
    _drunkins_set_process_config($job_id, 'items', array());
  }

  unset($job_context['drunkins_process_more']);

  // Log before start() (at debug level). Logging immediately after start() is
  // not done explicitly, unless 'enqueue_by_fetcher' is set: we usually wait
  // until items are queued.
  _drunkins_log($job_id, 'Calling start() for job "@job".', array('@job' => $job_id), WATCHDOG_DEBUG);
  // 'process_settings' and 'process_runner_settings' are set. (If this is the
  // first run: the process info has been updated to match the process as if it
  // was already running, just before calling us.)
  $runner_settings = _drunkins_get_process_config($job_id);
  // Set memory limit. We will not reset it; this is left up to the caller after
  // items are queued.
  if (!empty($runner_settings['memory_limit_start']) || !empty($runner_settings['memory_limit'])) {
    _drunkins_set_memory_limit($job_id, !empty($runner_settings['memory_limit_start'])
      ? $runner_settings['memory_limit_start'] : $runner_settings['memory_limit']);
  }
  try {
    $items = $job->start($job_context);
    // Protect against hacks/errors in the job class.
    if (!is_array($job_context)) {
      // We haven't defined our own exception type yet. Maybe later, if there's
      // a need.
      throw new \Exception('The context has a wrong format after returning from start().');
    }
  }
  catch (\Exception $e) {
    // Do cleanup and rethrow.
    _drunkins_finish_process($job_id, $first_run
      ? array('target_state' => 'error_on_start', 'delete_items' => FALSE)
      : array('target_state' => 'error_on_restart'));
    throw $e;
  }

  // Remove items if they were present because we don't want to keep / store
  // them in process state. (Only necessary for non-batch mode.) We lose them if
  // there's an error between this point and queuing $items - but that isn't
  // worse than losing normal $items.
  unset($job_context['drunkins_override_fetch']);

  // Remember the number of times we've run, only if 'drunkins_process_more' was
  // set (now or in the previous run). A job shouldn't mess with the
  // 'drunkins_rerun_start' value, but if it does we reset it to 1.
  if (!empty($job_context['drunkins_rerun_start']) && is_int($job_context['drunkins_rerun_start'])) {
    $job_context['drunkins_rerun_start']++;
    if (empty($runner_settings['enqueue_by_fetcher'])
        && isset($job_context['drunkins_rerun_start_count'])
        && is_numeric($job_context['drunkins_rerun_start_count'])) {
      $job_context['drunkins_rerun_start_count'] += count($items);
    }
    else {
      // If 'enqueue_by_fetcher' or enqueueing extra items explicitly, we don't
      // know if items were dequeued by now.
      $job_context['drunkins_rerun_start_count'] = '?';
    }
  }
  elseif (!empty($job_context['drunkins_process_more'])) {
    $job_context['drunkins_rerun_start'] = 1;
    $job_context['drunkins_rerun_start_count'] = empty($runner_settings['enqueue_by_fetcher'])
      ? count($items) : DrupalQueue::get($job_id)->numberOfItems();
  }

  return $items;
}

/**
 * Enqueue items into a running process.
 *
 * This function is not suited for calling start() before a process is really
 * started because it does not change a process' recorded state to "running".
 * (It only changes process state to "cancelled" if a start() call fails.) So at
 * the first call during a process' lifetime, $items will probably be passed
 * (because the rest of the code is set up such that starting a job also means
 * getting items returned), though the function does not depend on this.
 *
 * If the 'progressive' runner setting is explicitly set to FALSE, start() may
 * be called (repeatedly) during this one function call, to enqueue more items,
 * so the 'start phase' of the process always gets handled in one call.
 *
 * If 'progressive' is true/unset, the caller is expected to handle situations
 * where the job's start() signifies more items should be fetched, i.e. to call
 * this function repeatedly. Since the process status then is already "running",
 * the $items argument may be left empty so that start() is run by this function
 * before enqueueing its resulting items.
 *
 * If a start() call fails, the whole process will be canceled (as opposed to
 * just skipping the result of this one call).
 *
 * @param string $job_id
 *   Job ID, also used as machine name of the queue.
 * @param array $items
 *   (optional) $items to enqueue. If passed, the job's start() method will be
 *   run to fetch them, before queueing them.
 * @param array $override_runner_settings
 *   (optional) A specific subset of runner settings which can override the
 *   running process' settings for this call: 'logging', 'log_level' and
 *   'add_last_marker'. 'add_last_marker' has a default value of FALSE, meaning
 *   its value in the running process' settings is always ignored. Setting this
 *   argument implies items are added from an 'external' source and start() is
 *   not called to fetch them. start() may be called for preprocessing the
 *   passed items (depending on process settings), however the settings passed
 *   in this argument will not have effect during the call to
 *   _drunkins_run_start(); the running process' settings will be used there.
 *
 * @return bool
 *   TRUE if items were successfully fetched and queued (also if the number of
 *   items is 0); FALSE if the process state means this function should not be
 *   called or for error thrown during start() call.
 */
function _drunkins_enqueue_items($job_id, array $items = NULL, array $override_runner_settings = NULL) {
  $info = drunkins_job_status_info($job_id);
  if ($info['running'] != 3 || $info['method'] === 'batch') {
    _drunkins_log($job_id, "_drunkins_enqueue_items() called for job \"@job\" which is not running or is a 'batch' job.", array('@job' => $job_id), WATCHDOG_ERROR);
    return FALSE;
  }

  $context = &drunkins_current_context($job_id);

  $runner_settings = _drunkins_get_process_config($job_id);
  if (isset($override_runner_settings)) {
    if (empty($info['can_add_items_while_running'])) {
      _drunkins_log($job_id, 'Job "@job" does not support adding extra items when already running.', array('@job' => $job_id), WATCHDOG_ERROR);
      return FALSE;
    }

    // Restrict the override settings to specific ones. In particular, we are
    // disallowing the use of 'enqueue_by_fetcher' - because 1) it's important
    // that we follow the value used in _drunkins_run_start(), and that function
    // cannot use our override value... and 2) changing it from the running
    // process' value has no known usefulness and is potentially confusing. Also
    // ignore 'add_last_marker' from the running process.
    unset($runner_settings['add_last_marker']);
    $filter = array_flip(array('logging', 'log_level', 'add_last_marker'));
    $runner_settings = array_intersect_key($override_runner_settings, $filter) + $runner_settings;
    $run_start = !empty($info['enqueue_by_fetcher']) || $info['can_add_items_while_running'] === 'pass_through_start' || $info['can_add_items_while_running'] === 'pass_through_start_by_default';
    if ($run_start) {
      // Pass $items through start() for preprocessing. (We assume no new items
      // will be fetched in start().)
      if (empty($info['can_override_fetch'])) {
        if (!empty($info['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, "Job \"@job\" has no 'can_override_fetch' capability so it cannot handle the 'enqueue_by_fetcher' setting.", array('@job' => $job_id), WATCHDOG_ERROR);
        }
        else {
          _drunkins_log($job_id, "Job \"@job\" has inconsistent capabilities; items must be passed through start() but the job is not marked as being able to handle that (i.e. 'can_override_fetch' capability is not set).", array('@job' => $job_id), WATCHDOG_ERROR);
        }
        return FALSE;
      }
    }
    $can_loop = FALSE;

    // _drunkins_run_start() keeps the number of times start() was run, only if
    // specifically marked as 'start() needs rerun'. We will use the same
    // counter for logging purposes, regardless whether we actually run start()
    // or enqueue items directly.
    if (!isset($context['drunkins_rerun_start'])) {
      // Apparently start() was only run once and didn't need rerun. (Note that
      // when starting a new job from 'external' enqueued items, it also does
      // not get set; the 'else{}' block below gets executed instead.)
      // Set to 1 now.
      $context['drunkins_rerun_start'] = 1;
    }
    if (!$run_start) {
      // _drunkins_run_start() won't be executed so we'll do this ourselves:
      $context['drunkins_rerun_start']++;
    }
  }
  else {
    // 'first call' isn't always accurate; it's also TRUE when queueing extra
    // items received at finish().
    $run_start = !isset($items);
    if ($run_start && empty($context['drunkins_process_more'])) {
      // This process is not in a state where start() needs to be re-run, so
      // this function should not have been called.
      return FALSE;
    }

    // If this is a progressive job then we exit always; in that case if
    // $context['drunkins_process_more'] is nonempty, the next cron run / call
    // to the 'start again callback' should pick things up.
    // If this is the start of a non-progressive job, we might run again based
    // on $context['drunkins_process_more']. In practice this can only be the
    // case for a call where items were fed into this function; calling this
    // without $items for a process marked as 'non-progressive' is weird, given
    // the setup of the rest of the code.
    $can_loop = isset($runner_settings['progressive']) && empty($runner_settings['progressive']);
  }
  // If we run start() ourselves in the iteration below, we also want to reset
  // the memory limit (which _drunkins_run_start() can set, but does not reset
  // because the items are still in memory). If on the other hand the caller
  // has ran _drunkins_run_start() by itself (regardless whether start() is
  // called again if the job is progressive), we leave it up to the caller to
  // reset it. And obviously if we don't run start() at all, we don't reset it.
  $reset_memory_limit = $run_start;

  do {
    if ($run_start) {

      if (isset($override_runner_settings)) {
        // Prepare for adding extra items.
        _drunkins_set_process_config($job_id, 'items', $items);
        if (!empty($runner_settings['enqueue_by_fetcher'])) {
          $count = count($items);
        }
      }

      try {
        $items = _drunkins_run_start($job_id, $context);
      }
      catch (\Exception $e) {
        // Process failed; is already canceled, queue cleaned up.
        _drunkins_log($job_id, 'Rerunning start for job "@job" threw exception; deleting all previously gathered items from the job. Exception: @error', array('@job' => $job_id, '@error' => $e->getMessage()), WATCHDOG_ERROR);
        return FALSE;
      }
    }

    if ($items && empty($runner_settings['enqueue_by_fetcher'])) {
      $add_last_marker = empty($context['drunkins_process_more']) && !empty($runner_settings['add_last_marker']);
      $queue = DrupalQueue::get($job_id);
      $nr = count($items);
      $i = 1;
      foreach ($items as $item) {
        // As the worker callback is the same for all Drunkins jobs, and
        // it can have only one argument, we need to include an ID for the
        // job (the queue name) _in_ the item.
        $qitem = array('data' => $item, 'queue_name' => $job_id);
        if ($add_last_marker && $i++ == $nr) {
          // This is just a hint to help finish the job as soon as possible, if
          // this item is processed last in the queue; there is no guarantee
          // that it is actually processed last. (There is a >99.9% chance of it
          // in most environments, except when exceptions are thrown during
          // processItem() calls.)
          $qitem['last'] = TRUE;
        }
        $queue->createItem($qitem);
      }
    }

    // Do logging; the message depends on a lot of factors (starting the
    // process; adding extra items externally; needing to add more; knowing
    // how many items we actually enqueued)

    if (empty($runner_settings['enqueue_by_fetcher'])) {
      $count = count($items);
    }
    elseif (!isset($override_runner_settings)) {
      $count = DrupalQueue::get($job_id)->numberOfItems();
    }
    // If isset($override_runner_settings) && !empty($runner_settings['enqueue_by_fetcher']),
    // $count was already set (because $items are destroyed now BUT we knew the
    // probable amount of items queued).

    if (!empty($context['drunkins_process_more'])) {
      if (isset($override_runner_settings)) {
        _drunkins_log($job_id, "While adding extra items to running job \"job_id\" and passing those though start(), the 'drunkins_process_more' context value was set. This is illegal and will be ignored.",
          array('@job' => $job_id), WATCHDOG_WARNING);
        unset($context['drunkins_process_more']);
      }
      elseif (empty($runner_settings['enqueue_by_fetcher'])) {
        _drunkins_log($job_id, 'Fetched @count items for job "@job" (total @total in @rerun_count iterations); trying to fetch more...',
          array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count, '@total' => $context['drunkins_rerun_start_count'], '@job' => $job_id), WATCHDOG_INFO);
      }
      else {
        _drunkins_log($job_id, 'Fetched items for job "@job" (total -at least- @total in @rerun_count iterations); trying to fetch more...',
          array('@rerun_count' => $context['drunkins_rerun_start'], '@total' => $count, '@job' => $job_id), WATCHDOG_INFO);
      }
    }
    if (empty($context['drunkins_process_more'])) {
      if (isset($override_runner_settings)) {
        // 'drunkins_rerun_start' is the number of times start() has been run.
        // So the first time we enqueue extra items to a running process is
        // immediately marked '2' - or higher if start() was run multiple times.
        // This is the best / most consistent we can do.
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Added @count items to job "@job" from an external source (@rerun_count). ', array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Added (probably) @count items to job "@job" from an external source (enqueued by the job itself).', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
      elseif (!empty($context['drunkins_rerun_start'])) {
        // It's the last time this 'start again callback' is called. Part of
        // the items may or may not already have been processed.
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Fetched @count items for job "@job" (total @total in @rerun_count iterations); starting job.',
            array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count, '@total' => $context['drunkins_rerun_start_count'], '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Fetched items for job "@job" (total -at least- @total in @rerun_count iterations); starting job.',
            array('@rerun_count' => $context['drunkins_rerun_start'], '@total' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
      elseif (_drunkins_get_process_config($job_id, 'items') !== FALSE) {
        // We're adding items for this process for the first time, so apparently
        // the items were added now. (The order of the outer if()s is such that
        // this would also still be true if above condition stayed true over
        // several HTTP requests, even though it doesn't at the moment.)
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Starting job "@job"; added @count items from an external source.', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Starting job "@job"; added (probably) @count items from an external source (enqueued by the job itself).', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
      else {
        // This is the 'normal' situation where we are supposedly not called
        // from a 'start again callback' and $run_start was/will stay FALSE.
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Starting job "@job"; @count items to process', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Starting job "@job"; @count items to process (enqueued by the job itself).', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
    }

    $run_start = !empty($context['drunkins_process_more']) && $can_loop;

  } while ($run_start);

  if ($reset_memory_limit) {
    _drunkins_set_memory_limit($job_id);
  }

  return TRUE;
}

/**
 * Update job info in database and status info, to say a process has finished.
 *
 * Optionally deletes the queue / batch(set) / items.
 *
 * Supported options which influence behavior: (Even though the keys are long,
 * the only use is writing self documenting internal code; we do not
 * expect/want these options to be used by outside code.)
 *   'target_state' =>       string: can be 'manual_reset' / 'error_on_restart'
 *      / 'error_on_start' / 'empty_start' / 'error_on_finish'
 *      (If otherwise: target state will be 'normally finished'.
 *   'log_unknown_state'           boolean; default TRUE
 *   'delete_items'                boolean; default TRUE
 *   'job_may_restart_immediately' boolean
 *
 * (Note 'error_on_finish' really means 'stuck', not 'finished' and should
 * always be accompanied by delete_items=FALSE. That is: this function is not
 * only to "set a process to finished", but to "change a process from running".)
 */
function _drunkins_finish_process($job_id, $options = array()) {

  // Get updateable status info. (In some cases, like manual reset, this info
  // may not even be compiled yet and we end up wasting time. But at least this
  // way we are sure that the internally cached info is not stale.)
  $job_info = &drunkins_job_status_info($job_id);

  // Compile the database fields to update.
  $fields = array();
  switch (isset($options['target_state']) ? $options['target_state'] : '') {
    // See drunkins_job_status_info() function comments for values.
    case 'manual_reset':
      // This is also the 'default' for true-ish non-string values.
      $fields['run_info'] = 5;
      break;
    case 'error_on_start':
    case 'error_on_restart':
      $fields['run_info'] = 4;
      break;
    case 'error_on_finish':
      if ($job_info['method'] !== 'queue') {
        // There is no reason that this could happen except a code error.
        _drunkins_log($job_id, 'Trying to set job "@job"\'s target stage "@target_state" but this does not exist for @method processes. The module code must be reviewed.', array('@job' => $job_id, '@target_state', 'error_on_finish', '@method' => $job_info['method']), WATCHDOG_CRITICAL);
        // Not setting $fields['run_info'] means we don't update state but do
        // delete items.
        $options['delete_items'] = TRUE;
      }
      else {
        $fields['run_info'] = 3;
      }
      break;
    case 'empty_start':
    default:
      // Mark as normally finished.
      $fields['run_info'] = 0;
  }
  // Check if the info says that the process is actually running (or if we are
  // resetting a queue job it may also be in a 'stuck' state).
  // Otherwise, log the inconsistency and exit.
  if (!isset($job_info['run_info'])
      || !($job_info['run_info'] == 0 && $job_info['started'] > $job_info['finished']
           || $job_info['run_info'] == 1 || $job_info['run_info'] == 2
           || $job_info['run_info'] == 3 && isset($options['target_state']) && $options['target_state'] === 'manual_reset'
  )) {
    // We don't know what to do on failure, or what kind of sensible message to
    // give to the end user. This isn't expected to have a huge effect, though
    // it's a sign of stale code. So log only to watchdog, a warning that is
    // only readable by the programmer.
    // (This function is not supposed to be called by outside code.)
    if (!isset($options['log_unknown_state']) || !empty($options['log_unknown_state'])) {
      watchdog('drunkins', "Unknown process 'run_info' state \"@strange_state\" while calling _drunkins_finish_process() to set run_info=@new_info. The module code should be reviewed.",
        array('@strange_state' => $job_info['run_info'], '@new_info' => $fields['run_info']), WATCHDOG_WARNING);
    }
  }
  elseif (isset($fields['run_info'])) {
    // Compile the database fields to update.
    if (isset($options['target_state']) && $options['target_state'] === 'manual_reset') {
      // We don't want to leave a big blob of context data behind in an unknown
      // state. We may keep it in case the job was normally finished (or in case
      // of error, this could be useful for inspection?) but in this case
      // there's a higher chance of 'dangerous' data leaking.
      $info = db_query("SELECT data FROM {drunkins_process} WHERE queue_name = :qn",array(':qn' => $job_id))
        ->fetchAssoc();
      $data = unserialize($info['data']);
      unset($data['context']);
      $fields['data'] = serialize($data);
    }

    // 'error_on_finish' is not really 'finished' so don't change the time.
    if (!isset($options['target_state']) || $options['target_state'] !== 'error_on_finish') {
      $fields['finished'] = time();
      // If a new process could start immediately after this function exits, we
      // could have an issue when 'started' gets set to the current time
      // because 'started == finished' means a job is not running. So make sure
      // that 'finished' will be smaller than that theoretical start time by
      // decrementing it.
      // @todo while OK in practice, this still seems flakey. Maybe change our assumption elsewhere in the code, that only 'started > finished' means a job is running?
      if (!empty($options['job_may_restart_immediately'])) {
        if ($job_info['started'] == $fields['finished']) {
          // Theoretical edge case: by decreasing 'finished', the start time
          // would become greater so the process would seem to be running. Fix.
          $fields['started'] = $fields['finished'] - 1;
        }
        $fields['finished']--;
      }
    }

    db_update('drunkins_process')
      ->fields($fields)
      ->condition('queue_name', $job_id)
      ->execute();

    // Make sure the cached 'job info' is up to date too, to prevent obscure
    // bugs. (At least drunkins_cron_info() needs up to date 'running' info.)
    $job_info = array_merge($job_info, $fields);
    _drunkins_calculate_dynamic_info($job_info, $job_id);
  }

  // Remove the class instance so it will be reinstantiated if a new job
  // with the same ID gets started. In principle this is unnecessary
  // because the instance should not hold any state information, so we
  // should be able to reuse it... But we (arbitrarily) draw a line at
  // reusing class instances over different processes, to prevent hard to
  // trace bugs from class implementations that do not follow the rules.
  // (Bugs occurring during the lifetime of one process are supposedly less
  // hard to spot.)
  unset($job_info['process_instance']);

  // Delete items at the end; this is the expensive/least predictable operation.
  // It never makes sense to not delete items. However since this is an internal
  // function, and we in theory don't know how expensive the queue mechanism is,
  // a call can choose to not delete them if it knows they're not there.
  // (At this point, the drunkins cron-queue code would discard any queued
  // items it received without context. But purging the queue is neater.)
  if (!isset($options['delete_items']) || !empty($options['delete_items'])) {
    DrupalQueue::get($job_id)->deleteQueue();
  }
}

/**
 * Logging function which decides what 'backend' to use.
 *
 * This will be replaced by a logger class
 *
 * $override_runner_settings is for those rare cases where we have specific
 * logger settings but don't want to overwrite settings for the process.
 */
function _drunkins_log($job_id, $message, array $variables, $severity, $override_runner_settings = array()) {

  $runner_settings = $override_runner_settings + _drunkins_get_process_config($job_id);
  if ($severity <= (isset($runner_settings['log_level']) ? $runner_settings['log_level'] : WATCHDOG_INFO)) {

    $last_log = &drupal_static('drunkins_last_log');
    $last_log = array($message, $variables, $severity);

    $logging = isset($runner_settings['logging']) ? $runner_settings['logging'] : DrunkinsJob::LOG_WATCHDOG;

    if ($logging & DrunkinsJob::LOG_SCREEN) {
      $type = 'status';
      // NOTICE maps to 'status'; it would not stand out if we printed a lot of
      // status messages, but this function is expected to be used mostly by
      // warnings/errors - so this way a NOTICE will look different from a
      // WARNING on screen.
      switch ($severity) {
        case WATCHDOG_WARNING:
          $type = 'warning';
          break;
        case WATCHDOG_ERROR:
        case WATCHDOG_CRITICAL:
        case WATCHDOG_ALERT:
        case WATCHDOG_EMERGENCY:
          $type = 'error';
      }
      drupal_set_message(t($message, $variables), $type);
    }
    if ($logging & DrunkinsJob::LOG_WATCHDOG) {
      watchdog('drunkins', $message, $variables, $severity);
    }
    if ($logging & DrunkinsJob::LOG_DRUSH) {
      if (DRUSH_MAJOR_VERSION < 7) {
        // _drush_log() ignores 'info' (which is our default log severity) and
        // 'notice' if not in verbose mode - and then they are printed among
        // lots of other Drush logs. Alternatively, we can give severity
        // 'status' and then all messages get a green "[status]" indicator.
        // (Which can again be suppressed by 'quiet'.) Not sure if all logs
        // should be 'green' but we do want the "X items queued" message.
        $type = 'info';
        switch ($severity) {
          case WATCHDOG_INFO:
          case WATCHDOG_NOTICE:
            $type = 'status';
            break;
          case WATCHDOG_WARNING:
            // See just below.
            if ($this->settings['logging'] & DrunkinsJob::LOG_WATCHDOG) {
              return;
            }
            $type = 'warning';
            break;
          case WATCHDOG_ERROR:
          case WATCHDOG_CRITICAL:
          case WATCHDOG_ALERT:
          case WATCHDOG_EMERGENCY:
            // Watchdog already prints to the screen when Drush executes;
            // don't output duplicate logs.
            // @todo check what exact drush context this depends on.
            // @todo check drush >= 7.
            if ($logging & DrunkinsJob::LOG_WATCHDOG) {
              // This only works if LOG_DRUSH is the last if() block...
              return;
            }
            $type = 'error';
        }
        $severity = $type;
      }
      drush_log(t($message, $variables), $severity);
    }
  }
}

/**
 * Runs a queue process in a 'non-progressive' way, i.e. processes all items
 * in this one function call. Please see comments at drunkins_start_job() for
 * caveats; the concept of Drunkins jobs is not designed for doing this.
 *
 * There is no return value; drunkins_last_log() will contain the message
 * / context from finish(). (We assume that drunkins_last_log() at this moment
 * returns a value which does not contain context, so that a succesfully-fully
 * run process can be distinguished by that.)
 *
 * There's no reason for this to be a separate function except
 * drunkins_start_job() gets too long.
 *
 * @param $job_id
 *   Job ID / machine name for the the queue.
 * @param string $method
 *   Method used for running this job.
 *
 * @see drunkins_start_job()
 */
function _drunkins_process_queued_items($job_id, $method) {

  // Initialize last-log; we'll check that to see if the process finished.
  // It's near-impossible that it has 4 values now, but:
  $last_log = &drupal_static('drunkins_last_log');
  unset($last_log[3]);

  // Get pointer into 'live' context. We don't want to change anything in it.
  $job_context = &drunkins_current_context($job_id);

  $runner_settings = _drunkins_get_process_config($job_id);
  $progress_interval = !empty($runner_settings['update_progress_interval']) ? $runner_settings['update_progress_interval'] : 50;

  $queue = DrupalQueue::get($job_id);
  while (TRUE) {

    // Simulate a drupal_cron_run() that doesn't check execution time.
    while ($item = $queue->claimItem()) {
      _drunkins_cron_process_queue_item($item->data);
      $queue->deleteItem($item);

      if ($progress_interval && isset($job_context['drunkins_progress']) && is_numeric($job_context['drunkins_progress'])
          && ($job_context['drunkins_progress'] % $progress_interval) == 0 && $method === 'drush' && !drush_get_context('DRUSH_QUIET')) {
        // Show update progress. This is not a normal log message; it should
        // only be shown on the display for some methods - drush for now. We
        // will consider a drush progressbar library when we start using
        // composer.
        drush_print(t("processed @nr items\r", array('@nr' => $job_context['drunkins_progress'])), 0, STDERR, FALSE);
      }
    }

    // Doublecheck whether claimItem() failed on us.
    if ($queue->numberOfItems()) {
      _drunkins_log($job_id, 'While running job "@job": claimItem() returned false with items still queued. Aborting further processing.', array('@job' => $job_id), WATCHDOG_ERROR);
      // In the odd case that $return is still filled (which means the job
      // finished properly even though now there are suddenly items in the queue
      // which is only possible in some strange race condition), just return it.
      break;
    }

    if (isset($last_log[3])) {
      // 'last' item triggered a call to finish(); normal exit.
      break;
    }

    // 'last' item did not trigger a call to finish(); do it now.
    _drunkins_queue_process_check_finish($job_id);
    // This may cause the queue to be re-filled. If so, jump back in the loop.
    if ($queue->numberOfItems()) {
      continue;
    }

    if (!isset($last_log[3])) {
      $log_source = array();
      if ($runner_settings['logging'] & DrunkinsJob::LOG_SCREEN) {
        $log_source[] = 'on the screen';
      }
      if ($runner_settings['logging'] & DrunkinsJob::LOG_WATCHDOG) {
        $log_source[] = 'in watchdog logs';
      }
      _drunkins_log($job_id, "While running job \"@job\": finish() did not run correctly. See earlier messages !source for details.",
        array('@job' => $job_id, '!source' => implode('/', $log_source)), WATCHDOG_ERROR);
    }
    break;
  }

}

/**
 * Worker callback (1 item) for cron queue process, as called from .module.
 */
function _drunkins_cron_process_queue_item(array $qitem) {

  if (isset($qitem['queue_name']) && $qitem['queue_name'] !== '') {

    // If we end up here (i.e. there are queued items) but the job status is
    // 'not running' or context is empty, then discard items; we can't run
    // reliably without this info. This means that deleting a row in
    // {drunkins_process} is equivalent to canceling the running job.

    $context = &drunkins_current_context($qitem['queue_name']);
    if ($context) {
      $job = drunkins_get_job($qitem['queue_name']);
      if ($job) {
        // Process one item.
        // 'progress' gets incremented every time before an item is processed.
        // (This means it's not accurate in case of parallel processing, but
        // then again, this whole module can't deal with parallel processing
        // because $context can lose info too.)
        $context['drunkins_progress']++;
        try {
          $job->processItem($qitem['data'], $context);
        }
        catch (\Exception $e) {
          $runner_settings = _drunkins_get_process_config($qitem['queue_name']);
          if (isset($runner_settings['handle_process_exception'])
              && $runner_settings['handle_process_exception'] === 'discard') {
            // Do nothing except accounting. The item will be dequeued.
            _drunkins_log($qitem['queue_name'], 'Processing "@job" item failed: @message', array('@job' => $qitem['queue_name'], '@message' => $e->getMessage()), WATCHDOG_ERROR);
            $context['drunkins_exception_count']++;
            // @todo do something with this number? Show it in UI?
          }
          else {
            // Default value: 'fallthrough'. (Other values should have been
            // caught before process start.)
            throw $e;
          }
        }
      }
    }

    if (!empty($qitem['last'])) {
      if ($context) {
        // 'last' means it was the last item queued for this job originally,
        // which means it's _probably_ the last item processed. This is the only
        // case where we execute DrupalQueue::get($queue_name)->numberOfItems()
        // to check if we can finish the job within this worker callback; it is
        // considered to be too expensive to call inside every callback.
        // If the job is not finished now, it will get finished at the beginning
        // of the next cron run after all items have already been processed.
        // (This means that a system which calls its own worker callbacks, will
        // also have to make sure that the job gets finished, after all items
        // are processed.)
        // The last item will automatically be removed after returning from this
        // worker callback, so a job's finish() will be called _before_ the
        // queue is empty. This should not matter (finish() has no knowledge of
        // the queue mechanism).
        _drunkins_queue_process_check_finish($qitem['queue_name'], array('last_item_still_in_queue' => TRUE));
      }
      else {
        // $context was deleted at some point and probably all queued items fell
        // through silently. Log a warning for this last one.
        _drunkins_log($qitem['queue_name'], 'Completed queued job "@job": @message', array(
            '@job' => $qitem['queue_name'],
            '@message' => t('Several items were discarded because context / run status was invalid.')
          ), WATCHDOG_WARNING);
      }
    }
  }
}

/**
 * Check if a queue-mode process* which is marked 'running' has processed all
 * items, and if it has, finish it. Note a possible effect of this is that the
 * queue is repopulated with extra items, returned by finish().
 *
 * *meaning: a process which uses the Drupal Queue mechanism to process its
 * items. This can be more than just the 'queue' method.
 *
 * Supported options which influence behavior: (Even though the keys are long,
 * the only use is writing self documenting internal code; we do not
 * expect/want these options to be used by outside code.)
 *  'last_item_still_in_queue' (boolean)
 *  ...any options supported by _drunkins_finish_process(), a.o.:
 *    'job_may_restart_immediately' (boolean)
 */
function _drunkins_queue_process_check_finish($job_id, $options = array()) {

  // If job isn't running, silently skip it.
  $info = &drunkins_job_status_info($job_id);
  if ($info['running'] != 3 || $info['method'] === 'batch') {
    // We may decide later to delete this if(); it really should never happen.
    _drunkins_log($job_id, 'Current method @method or run_info code @run_info of job "@job" is illegal when calling _drunkins_queue_process_check_finish(). This indicates a code error.', array('@method' => $info['method'], '@run_info' => $info['run_info'], '@job' => $job_id), WATCHDOG_CRITICAL);
  }
  // If 'drunkins_process_more' is set, skip it. If context is empty (which
  // should never happen) then 'finish' it too but log an error.
  elseif (empty($info['process_context']['drunkins_process_more'])) {
    // Recheck if the number of items is really 0 (i.e. the job is really
    // finished) and exit silently if it isn't. (Some callers use this feature.)
    // A potential race condition exists because an item may be added just
    // after checking that the queue is empty. This item will be discarded. If
    // you're worried about this, then use DrunkinsOrderedQueue.
    // @todo implement DrunkinsOrderedQueue which does not accept new items once empty - also should process items in exact order.
    $items = DrupalQueue::get($job_id)->numberOfItems();
    if (!empty($options['last_item_still_in_queue']) ? $items < 2 : !$items) {

      $runner_settings = array();
      $context = &drunkins_current_context($job_id);
      if ($context) {
        $job = drunkins_get_job($job_id);
        if ($job) {
          // Log before finish() (at debug level).
          // @todo what do we exactly want to do immediately after, when finish() does not return a message anymore?
          _drunkins_log($job_id, 'Calling finish() for job "@job".', array('@job' => $job_id), WATCHDOG_DEBUG);
          $runner_settings = _drunkins_get_process_config($job_id);
          if (!empty($runner_settings['memory_limit_finish'])) {
            _drunkins_set_memory_limit($job_id, $runner_settings['memory_limit_finish']);
          }
          //@todo think: is there a chance of a race condition when another process is
          //   adding an item to the queue of this running job, while we are already
          //   running finish()? What to do about it?
          //@todo think: is there a chance of a race condition when we run
          //   _drunkins_finish_process() but the last queue item still needs to
          //   be dequeued (by the cron code)? If so, what to do about it?
          //   If not, maybe make a comment somewhere.
          try {
            $message = $job->finish($context);

            if (!empty($context['drunkins_process_more']) && is_array($context['drunkins_process_more'])) {
              // If we got any message back, don't just drop it; log/print it.
              if ($message) {
                _drunkins_log($job_id, $message, array(), WATCHDOG_INFO);
              }
            }
            else {
              _drunkins_log($job_id, 'Completed queued job "@job": @message', array('@job' => $job_id, '@message' => $message), WATCHDOG_INFO);
              // TODO display message/context/settings in drunkins UI/overview.

              // Last log was stored, but also store context as a signal that the
              // process was successfully finished.
              $log = &drupal_static('drunkins_last_log');
              $log = array($message, array(), WATCHDOG_INFO, $context);
            }
          }
          catch (Exception $e) {
            //@TODO if we make the items the return value, we can simplify a lot below because we cannot generate an exception AND more items.
            //  We'll also have to think about the other @todo below though.
            if (isset($runner_settings['handle_finish_exception'])
                && ($runner_settings['handle_finish_exception'] === 'continue'
                    || $runner_settings['handle_finish_exception'] === 'finish')) {
              // Log exception and finish normally (or continue on more items).
              _drunkins_log($job_id, 'Calling finish() for job "@job" failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
              if (!empty($job_context['drunkins_process_more']) && is_array($job_context['drunkins_process_more'])) {
                if ($info['handle_finish_exception']['continue_on_finish_exception'] === 'continue') {
                  _drunkins_log($job_context, '"@job" finish() generated @count more items to process: continuing despite exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
                }
                else {
                  _drunkins_log($job_id, '"@job" finish() generated @count more items to process: discarding them because of exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
                  unset($job_context['drunkins_process_more']);
                }
              }
            }
            else {
              // Default value: 'break'. (Other values should have been caught
              // before process start.)

              // This function is called in two situations:
              // - from a cron queue worker: we cannot rethrow the exception
              //   because the last item is still queued and we must let the
              //   caller dequeue it (so it will not get re-processed). We have
              //   to just log a message. (And then we could potentially just
              //   keep the process state "running" which would get us to the
              //   next situation some time later).
              // - from some other place that checks if the job is registered
              //   as status 'running' but has 0 items. At least one caller can
              //   potentially do this quite often: drunkins_cron_queue_info().
              //   Since we don't want to introduce a situation where a failing
              //   finish() is called repeatedly and indefinitely in the
              //   background... we need to introduce an extra 'error on finish'
              //   state (which either us or the caller should set).
              // For uniformity's sake, we won't throw and will move the process
              // to this state in both above cases. Note this 'error on finish'
              // state will only exist for queue processes because for batch,
              // there is no risk of constantly rerunning finish / the runner
              // will just throw an exception which the caller handles in a way
              // we like, if we keep the process status the same. (See comments
              // at drunkins_batch_finished().) Note also that there is no way
              // to rerun finish() for a process in 'error on finish' state
              // (unlike a standard interactive batch process where you can just
              // reload the page); a process can only be completely reset.
              // @todo make rerunning finish() possible, if we ever care.
              _drunkins_log($job_id, 'Calling finish() for job "@job" failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
              // We're not logging about any extra items; they *probably* will
              // be saved into context (if the PHP 'shutdown functions' still
              // run after an exception; haven't tested that). Potentially they
              // could be processed if the user reloads and triggers finish()
              // and that finishes without throwing an exception next time.
              // There is no facility for that at the moment, though.
              // @todo this completely changes if we make the extra items a return
              //   value instead of a subkey in $context. Think about this.

              $options['target_state'] = 'error_on_finish';
              $options['delete_items'] = FALSE;
            }
          }
        }
        else {
          // This should never happen.
          _drunkins_log($job_id, 'Failed to create class instance for job "@job" while trying to finish it!?', array('@job' => $job_id), WATCHDOG_ERROR);
          $options['target_state'] = 'error_on_finish';
          $options['delete_items'] = FALSE;
        }
      }
      else {
        _drunkins_log($job_id,
          '_drunkins_queue_process_check_finish() called without context for job "@job"; context was apparently missing from the database table?',
          array('@job' => $job_id, '@c' => $info['runner_settings']['class']),
          WATCHDOG_ERROR);
      }

      // @todo consider if we want to make this array (part of) the return
      // argument, instead of the message (if we want to get rid of that message as return argument anyway).
      if (!empty($context['drunkins_process_more']) && is_array($context['drunkins_process_more'])) {
        // @todo maybe make the 'executive decision' that status logs like the below
        // will not be done by the runner, and logging is the exclusive responsibility
        // of the job (except in clear errors)? Not sure; we should treat it the same as the 'N items received' from start().
        _drunkins_log($job_id,
          'finish() for job "@job" returned @count new items; continuing...',
          array('@job' => $job_id, '@count' => count($context['drunkins_process_more'])),
          WATCHDOG_INFO);
        // Enqueue items and continue. This also resets memory limit.
        _drunkins_enqueue_items($job_id, $context['drunkins_process_more']);
        // Unset 'drunkins_process_more' so hook_cron(_info)() won't assume
        // start() needs to be called again, and to free memory.
        unset($context['drunkins_process_more']);
        // Reset memory limit from the 'finish' one, if applicable. If there is
        // a general memory_limit setting, we will set that rather than the
        // initial one, since our caller may continue processing items.
        if (!empty($runner_settings['memory_limit_finish'])) {
          $initial_limit = _drunkins_get_initial_memory_limit();
          _drunkins_set_memory_limit($job_id, $initial_limit > $runner_settings['memory_limit'] ? $initial_limit : $runner_settings['memory_limit'], TRUE);
        }
      }
      else {
        _drunkins_finish_process($job_id, $options + array('delete_items' => FALSE));

        // drunkins_current_context() will still be called from
        // drunkins_flush_info(); empty static variable so it won't do anything.
        $context = array();
      }

    }
  }
}

/**
 * Returns a class instance for a job.
 *
 * Calls with and without the $settings argument serve a somewhat different
 * purpose. For $settings: see below. When not passing $settings: if the job is
 * not running, the returned instance is initialized with the standard settings
 * as defined in hook_queue_info(). If it is running, the instance is
 * initialized with the settings from the running process.
 * (That is: the settings which were passed to e.g. drunkins_start_*_job() to
 * start the process, plus the standard settings.)
 *
 * This function does nothing about starting jobs; it only gets instances for
 * the current job state. It shouldn't be as 'user facing' as it looks now. (In
 * fact it's only called from .run.inc and might be moved there.)
 *
 * @param string $job_id
 *   Job ID, also used as machine name of the queue for a cron/queue process
 * @param array $settings
 *   Settings influencing this specific job. When passing this argument, running
 *   jobs are ignored and the returned job instance will be initialized with
 *   $settings plus the standard settings from hook_queue_info(). This argument
 *   is passed only by code which is sure that a job is not running, or wants to
 *   use a job class outside of the regular 'process flow'.
 *
 * @return NULL|DrunkinsJob
 *   Class instance ready for use, or NULL if something went wrong.
 *
 * @todo investigate in how far we can get rid of the $settings argument, and if
 *   we can unify their behavior more with how 'runner settings' are currently done:
 *   store them in _drunkins_start_process() and never call this function before
 *   that time. Not sure how well that sits with the fringe cases where we need
 *   this function.
 */
function drunkins_get_job($job_id, array $settings = NULL) {

  $job_info = &drunkins_job_status_info($job_id);
  if ($job_info) {
    if (isset($job_info['process_instance']) && !isset($settings)) {
      return $job_info['process_instance'];
    }

    if ($job_info['running'] == 3 && !isset($settings)) {
      // Get settings from the running process.
      if (!isset($job_info['process_runner_settings']) || !is_array($job_info['process_runner_settings'])) {
        // If settings is not an array it is modified (by reference) by buggy
        // code. We won't run with buggy settings / need at least 'class' below.
        if ($job_info['method'] === 'batch') {
          // @todo we are not storing runner settings yet. Delete this when we do.
          $job_info['process_runner_settings'] = $job_info['runner_settings'];
        }
        else {
          return NULL;
        }
      }
      $runner_settings = $job_info['process_runner_settings'];

      if (!isset($job_info['process_settings']) || !is_array($job_info['process_settings'])) {
        return NULL;
      }
      $job_settings = $job_info['process_settings'];
    }
    else {
      // Don't get settings from the running process.

      if (!isset($job_info['runner_settings']) || !is_array($job_info['runner_settings'])) {
        // If settings is not an array it is defined wrongly, or modified (by
        // reference) by buggy code. We won't run with buggy settings and need
        // at least 'class' below.
        // @todo log?
        return NULL;
      }
      $runner_settings = $job_info['runner_settings'];

      if (!isset($job_info['settings'])) {
        $job_settings = array();
      }
      elseif (!is_array($job_info['settings'])) {
        // @todo log?
        return NULL;
      }
      else {
        $job_settings = $job_info['settings'];
      }
      if (isset($settings)) {
        $job_settings = array_merge($job_settings, $settings);
      }
    }

    if (!isset($runner_settings['class'])) {
      // @todo once we've implemented logger, see how to replace watchdog calls.
      watchdog('drunkins', "drunkins_job_status_info() has no 'class' definition for job \"@job\"",
        array('@job' => $job_id), WATCHDOG_ERROR);
    }
    elseif (!is_string($runner_settings['class'])) {
      watchdog('drunkins', "drunkins_job_status_info() has invalid 'class' definition for job \"@job\": @c",
        array('@job' => $job_id, '@c' => print_r($runner_settings['class'], TRUE)), WATCHDOG_ERROR);
    }
    else {
      // Add any 'runner settings' that must be propagated to the job
      // class because they are also supported there. (This is kind-of hackish
      // and dependent on hardcoded logic in default jobs.)
      // - 'fetcher_class' is copied into settings as 'drunkins_fetcher_class'.
      // - 'enqueue_by_fetcher' is copied into settings as
      //   'drunkins_enqueue_by_fetcher'; this is likely a temporary measure.
      // @todo to formally support this, we will need to inject something (maybe a Queue, maybe a Runner class)
      //   through context. At that time we might decide to move 'enqueue_by_fetcher' to context too...
      // - 'logging' is copied into settings as 'logging'.
      // Define an extra variable because we don't want to change the
      // 'process_runner_settings' which are stored in status info.
      $settings_not_stored_in_info = array();
      if (isset($runner_settings['fetcher_class']) && is_string($runner_settings['fetcher_class'])) {
        $settings_not_stored_in_info['drunkins_fetcher_class'] = $runner_settings['fetcher_class'];
      }
      if (!empty($runner_settings['enqueue_by_fetcher'])) {
        $settings_not_stored_in_info['enqueue_by_fetcher'] = $runner_settings['enqueue_by_fetcher'];
      }
      if (isset($runner_settings['logging'])) {
        $settings_not_stored_in_info['logging'] = $runner_settings['logging'];
      }
      // Also add job_id.
      $settings_not_stored_in_info['job_id'] = $job_id;
      try {
        $instance = new $runner_settings['class']($job_settings + $settings_not_stored_in_info, $job_id);
        // Cache the job instance if it's created for a running process.
        if (!isset($settings) && $job_info['running'] == 3) {
          $job_info['process_instance'] = $instance;
        }
        return $instance;
      }
      catch (Exception $e) {
        watchdog('drunkins', 'drunkins_get_job() failed to instantiate class @c for job "@job": @e',
          array('@job' => $job_id, '@c' => $runner_settings['class'], '@e' => $e->getMessage()), WATCHDOG_ERROR);
      }
    }
  }
}

/**
 * Returns a reference to a running non-batch process' active context.
 *
 * This is the (mandatory) way to obtain a reference to context data by code
 * that might alter the process context. This function takes care of 'lazily'
 * saving the context to persistent storage whenever a new process is switched
 * to or when the PHP process exits, and remembers current/up to date context
 * in the meantime.
 *
 * This is not supported for 'batch' method because that has its own way of
 * handling context.
 *
 * @param string $queue_name
 *   The ID of the job whose context must be returned; after the first call with
 *   this particular job id, the function will be responsible for saving its
 *   context at a later point. '' will return whatever context is 'active'
 *   (though there is no known practical use for this at the moment). The job
 *   must be running in queue mode; otherwise an empty array is returned.
 *
 * @return array
 *   The context data.
 */
function &drunkins_current_context($queue_name) {
  // @todo consider storing only the settings overrides in the database;
  //   not the queue info. Better for security/privacy, e.g. no passwords stored
  //   but on the other hand that could mean that if a job config gets changed,
  //   its settings/runner settings change too. Think about this. Config option?

  static $current_queue = '';

  // Write the current context to a persistent cache, before starting to
  // work on a new queue.
  // We use our dedicated table for this:
  // - no variable_set(); this clears caches;
  // - no ctools_object_cache_set because that depends on session info.
  // Note we might also wait until the PHP process ends and save all changed
  // processes' info to persistent cache at shutdown. However, writing whenever
  // another process 'becomes current' seems like a reasonable tradeoff between
  // speed and safety of not losing data.
  if ($queue_name !== $current_queue && $queue_name !== '') {

    if ($current_queue === '') {
      // This is the first queue item from the first job in this HTTP request.
      // We don't need to write an existing cache yet, but make sure this is
      // done on shutdown.
      drupal_register_shutdown_function('drunkins_flush_info');
    }
    else {
      $info = drunkins_job_status_info($current_queue);

      // If context is emptied somehow, we assume the job doesn't exist anymore
      // and the database info has already been updated.
      if (!empty($info['process_context'])) {

        // As a counterpart to _drunkins_queue_info_alter(), collect only known
        // properties that should be stored in the 'data' database field:
        $data = array(
          'context' => $info['process_context'],
          'settings' => $info['process_settings'],
        );
        if (!empty($info['process_runner_settings'])) {
          $data['runner_settings'] = $info['process_runner_settings'];
        }
        // Do a simple update query; if another process just deleted our info,
        // the context will not be saved and processing will stop.
        $updated = db_update('drunkins_process')
          ->fields(array('data' => serialize($data)))
          ->condition('queue_name', $current_queue)
          ->execute();
        if (!$updated) {
          // One possible cause of the update failing is the serialized data
          // being too big. Set empty context; if a database row indeed still
          // exists, this will cause all items to be discarded, which could be
          // better than processing them with stale context.
          // This fact will be logged in _drunkins_cron_process_queue_item().
          db_update('drunkins_process')
            ->fields(array('data' => serialize(array())))
            ->condition('queue_name', $current_queue)
            ->execute();
          // Below is a possibly-temporary measure using a variable that is not
          // exposed anywhere. If set, e-mail someone about the failure because
          // this is potentially serious; if context data is lost quietly then
          // (depending on the use case) processes running in the background
          // could stop running quietly. For being warned about this with a
          // (very basic text) e-mail, set an e-mail address here. This pairs
          // with the check in drunkins_flush_info().
          if ($mailto = variable_get('drunkins_fatal_error_mailto', '')) {
            // Setting '..._include_context' variable is security sensitive.
            _drunkins_mail_fatal_error($mailto, "Failed to save context for $queue_name; the process will likely abort."
              . (variable_get('drunkins_fatal_error_mail_include_context', FALSE)
                ? serialize($data) : ''));
          }
        }
      }
    }
    $current_queue = $queue_name;
  }

  // Get reference to full status info so we can manipulate the static array.
  $info = &drunkins_job_status_info($current_queue);

  if (isset($info['process_context']) && $info['running'] == 3 && $info['method'] !== 'batch') {
    return $info['process_context'];
  }
  // Return (a 'dead end' reference into) an empty array if $current_queue info
  // does not exist, has no context, or the process is not running or not in
  // queue mode. This way, calls to this function for nonexistent jobs cannot
  // pollute 'job status info'. (Unlike calls to drunkins_job_status_info().)
  $return = array();
  return $return;
}

/**
 * Shutdown function: Writes the current context to a persistent cache.
 *
 * @see drunkins_current_context()
 */
function drunkins_flush_info() {
  // This function is only registered from drunkins_current_context() so we
  // know there actually is a 'current context', and we don't need to put
  // this function in the .module.

  // We can use this to detect fatal errors like Out Of Memory too. In that
  // sense it pairs with the mail function in drunkins_current_context(), which
  // would not be triggered if OOM was encountered while serializing context.
  // It's not water tight, though; fatal errors that happen when _we_ are
  // calling drunkins_current_context() will not be caught.
  $error = error_get_last();
  if ($error && $error['type'] ^ E_ERROR && ($mailto = variable_get('drunkins_fatal_error_mailto', ''))) {
    _drunkins_mail_fatal_error($mailto, "File: $error[file]\nLine: $error[line]\nMessage: $error[message]");
  }

  // Write current context as an effect of switching to a different one.
  // We assume there is no queue named "\t".
  drunkins_current_context("\t");

  // Dereference class instances. (Probably unnecessary but doesn't hurt.)
  $info = &drunkins_job_status_info();
  $info = array();
}

/**
 * Send simple mail about fatal errors, bypassing hooks/alters.
 *
 * It's not sure whether this functionality will stay, in its current form. For
 * now it's dubbed "stopgap that seemed useful, but has not been tested much
 * in practice".
 */
function _drunkins_mail_fatal_error($to, $info) {
  // Pick the code from drupal_mail() that allows for configuration of the
  // mailer and has little chance of blowing up. Skip the alters and the
  // unneeded formatting.
  $default_from = variable_get('site_mail', ini_get('sendmail_from'));

  // This is what is left of drupal_mail() $message when you strip away the
  // stuff used by alters etc:
  $message = array(
    // 'from' is unecessary - except maybe for some obscure mailer class?
    'from'      => $default_from,
    'headers' => array(
      'From' => $default_from,
      'Sender' => $default_from,
      'Return-Path' => $default_from,
      'MIME-Version' => '1.0',
      'Content-Type' => 'text/plain; charset=UTF-8; format=flowed; delsp=yes',
      'Content-Transfer-Encoding' => '8Bit',
      'X-Mailer' => 'Drupal',
    ),
    'to'      => $to,
    'subject' => 'Drunkins encountered possibly-fatal error in ' . check_plain(variable_get('site_name', 'Drupal')),
    'body'    => "Info provided:\n\n$info",
  );

  $system = drupal_mail_system('drunkins', 'fatal_error');
  // DefaultMailSystem::format() does a LOT of proccessing, which can cause OOM
  // if $info is large. But since mail() also does at least a preg_replace(),
  // let's just do it. At least the contents will be safe. And if we encounter
  // an OOM it should be caught by the shutdown function which will then send a
  // much smaller mail.
  $message = $system->format($message);
  $system->mail($message);
}

/**
 * Sets or resets PHP memory limit.
 *
 * If we recorded that we already set the same (/ possibly a higher) limit with
 * a previous call, exit.
 *
 * @param string $job_id
 *  Job ID, used for logging. (Get rid of this parameter when writing runner.)
 * @param string $new_limit
 *   (Optional) limit to set. If not passed, an old limit is reset if that is
 *   known (and if it isn't known, the function exits without doing anything.)
 * @param bool $downsize
 *   (Optional) if TRUE, allow lowering of the current memory limit. If FALSE,
 *   only set if it is higher than the current limit. If $new_limit is empty,
 *   this parameter is ignored and lowering is allowed.
 *
 * @return int|bool
 *   If the limit could be set: the value of memory_limit before the call. FALSE
 *   if the new limit could not be set (because the given limit is lower than
 *   the current one, or because the ini_set() call failed).
 */
function _drunkins_set_memory_limit($job_id, $new_limit = '', $downsize = FALSE) {
  $initial_limit = &drupal_static(__FUNCTION__, 0);

  if (empty($initial_limit) && empty($new_limit)) {
    // We can't set an old initial value because we have not been called before.
    return FALSE;
  }

  if (empty($initial_limit) || !$downsize) {
    // Get current limit, to remember or to compare to $new_limit.
    $current_limit = parse_size(ini_get('memory_limit'));

    if (empty($initial_limit) ) {
      $initial_limit = $current_limit;
    }

    if ($new_limit) {
      if (!$downsize && parse_size($new_limit) <= $current_limit) {
        // Don't downsize.
        return FALSE;
      }
    }
    elseif ($initial_limit == $current_limit) {
      // Don't do ini_set if it's a no-op.
      return $current_limit;
    }
  }

  // Set new limit if provided; set initial limit otherwise.

  $old = @ini_set('memory_limit', $new_limit ? $new_limit : $initial_limit);

  if (!$old) {
    _drunkins_log($job_id, 'Failed setting memory limit to @mem', array('@mem' => $new_limit ? $new_limit : $initial_limit), WATCHDOG_WARNING);
  }

  return $old;
}

/**
 * Returns the PHP memory limit before any _drunkins_set_memory_limit() calls.
 *
 * @return int
 *   Initial memory limit; 0 if _drunkins_set_memory_limit() was never called.
 */
function _drunkins_get_initial_memory_limit() {
  return drupal_static('_drunkins_set_memory_limit', 0);
}

/**
 * Get configuration for currently running process.
 *
 * Internal function; doesn't have a lot of error checking.
 *
 * @param string $job_id
 * @param string $subsystem
 *   (optional) Subsystem / type of configuration:
 *   - 'runner' (default)
 *   - 'job'
 *   - 'items': returns the custom array of items to process in the job that is
 *     being started at the moment. FALSE if not set. $subset argument must be
 *     the default / not passed.
 * @param string $subset
 *   (optional) Subset of configuration:
 *   - 'running_process': options necessary for a running process. For subsystem
 *     'job', This may include a special key 'drunkins_start_items_overridden'
 *     which is set to TRUE if custom items were set for this job.
 *   - 'default': default configuration.
 *
 * @return array|bool
 */
function _drunkins_get_process_config($job_id, $subsystem = 'runner', $subset = 'running_process') {
  $info = drunkins_job_status_info($job_id);

  switch ("$subsystem.$subset") {
    case 'runner.running_process':
      // If 'process settings' not set yet, return the fixed configuration for
      // this job. (This makes it less necessary to guess whether one is
      // already allowed to get the "all" config / know when the custom config
      // is already set.)
      $config = !empty($info['process_runner_settings']) ? $info['process_runner_settings'] :
        (!empty($info['runner_settings']) ? $info['runner_settings'] : array());
      break;

    case 'runner.default':
      $config = !empty($info['runner_settings']) ? $info['runner_settings'] : array();
      break;

    case 'job.running_process':
      $config = !empty($info['process_settings']) ? $info['process_settings'] :
        (!empty($info['settings']) ? $info['settings'] : array());
      break;

    case 'job.default':
      $config = !empty($info['settings']) ? $info['settings'] : array();
      break;

    case 'items.running_process':
      $config = drupal_static('drunkins_set_process_items', FALSE);
      break;

    default:
      $config = FALSE;
  }

  return $config;
}

/**
 * Set configuration for currently running process.
 *
 * Configuration cannot be set incrementally with subsequent calls; if it is set
 * then previous 'dynamic' settings are erased. (They are just merged with any
 * 'static' configured settings.) This means we can run/finish a job and set up
 * another starting job with different settings, in the same call.
 *
 * @param string $job_id
 * @param string $subsystem
 *   (optional) Subsystem to set configuration for:
 *   - 'runner' (default).
 *   - 'job'.
 *   - 'items': custom set of items to start this job with. (Hint:
 *     clearing them when not needed anymore, might free memory.) The 'job'
 *     subsystem must not be set anymore after 'items' is set; if that is done
 *     then a special flag signifying that custom items are set, is lost. (It's
 *     not hugely important, but still it's nice to have.)
 *
 * @param array $custom_settings
 *   Array of settings. in case $subsystem == 'items': an array of items.
 */
function _drunkins_set_process_config($job_id, $subsystem, array $custom_settings) {
  $info = &drunkins_job_status_info($job_id);

  switch ($subsystem) {
    case '*runner_batch_logging_hack':
      // Set process runner settings to configured ones AND add screen logging.
      // @todo delete this once we store runner settings throughout the lifetime
      //   of batch processes.
      $custom_settings = array('logging' => DrunkinsJob::LOG_SCREEN | DrunkinsJob::LOG_WATCHDOG);
      // no break!
    case 'runner':
      if (isset($info['runner_settings']) && is_array($info['runner_settings'])) {
        $custom_settings += $info['runner_settings'];
      }
      $info['process_runner_settings'] = $custom_settings;
      break;

    case 'job':
      if (isset($info['settings']) && is_array($info['settings'])) {
        $custom_settings += $info['settings'];
      }
      $info['process_settings'] = $custom_settings;
      break;

    case 'items':
      $items = &drupal_static('drunkins_set_process_items');
      $items = $custom_settings;
      // Modify job settings to remember the items were overridden. This must
      // be done on set (not get) because otherwise this won't be saved in the
      // database. (We do not actually use this property anywhere but we might.)
      // This must be set after any 'job' setting, because modifying 'job' does
      // not keep this setting. (Which is on purpose; a new 'job' setting call
      // may pertain to a completely newly started process.)
      $info['process_settings']['drunkins_start_items_overridden'] = TRUE;
      break;
  }
}

/**
 * Alters queue info by adding info on processes.
 * Internal; meant only to be called from our hook_cron_queue_alter().
 *
 * @see drunkins_job_status_info()
 */
function _drunkins_queue_info_alter(array &$queues) {
  // Code note: drupal_alter('queue_info') may be called twice: independently,
  // and through drunkins_job_status_info(). (In D8: drupal_cron_run() does
  // this; in D7, only drunkins_overview().)

  $process_info = db_query("SELECT * FROM {drunkins_process}")->fetchAllAssoc('queue_name', PDO::FETCH_ASSOC);

  foreach ($queues as $queue_name => &$queue_info) {

    // Check data and merge 'process info' & calculated values. Only queues with
    // 'drunkins' set are important; others should never have 'process info'.
    if (!isset($queue_info['drunkins'])) {
      continue;
    }

    $info =& $queue_info['drunkins'];

    // Add 'process info' from database.

    if (isset($process_info[$queue_name])) {

      if ($process_info[$queue_name]['started'] > $process_info[$queue_name]['finished']) {
        // For running processes, merge only 'known' keys from the data column,
        // into 'process_*' properties:
        $process_data = unserialize($process_info[$queue_name]['data']);
        // If unserializing fails, _drunkins_cron_process_queue_item() will
        // discard all further queued items processed because context is empty.
        unset($process_info[$queue_name]['data']);

        if (isset($process_data['settings']) && is_array($process_data['settings'])) {
          $info['process_settings'] = $process_data['settings'];
        }
        if (isset($process_data['context']) && is_array($process_data['context'])) {
          $info['process_context'] = $process_data['context'];
        }
        if (isset($process_data['runner_settings']) && is_array($process_data['runner_settings'])) {
          $info['process_runner_settings'] = $process_data['runner_settings'];
        }
      }

      $info = array_merge($info, $process_info[$queue_name]);
    }

    // Calculate dynamic info.
    _drunkins_calculate_dynamic_info($info, $queue_name);
  }
  // For future re-coding, note $info / $queue_info still points inside $queues.
}

/**
 * Calculates dynamic values from other job info components; these are 'running'
 * plus optionally 'run_next' timestamp (which is only used by UI).
 *
 * @param array $info
 * @param string $job_id
 * @param bool $force_calc_next
 *   If TRUE, always calculate next cron-time if possible. By default it does
 *   not get calculated if the job is still running because it won't be checked
 *   by any process runner. Pass TRUE if you want to e.g. display it in a list.
 *
 * @see drunkins_job_status_info()
 */
function _drunkins_calculate_dynamic_info(array &$info, $job_id, $force_calc_next = FALSE) {
  // Just to be sure, protect against external sources setting this:
  unset($info['running']);
  unset($info['run_next']);

  if (!isset($info['run_info'])) {
    // No info (current/previous process run stored in database) available.
    // Calculate using only info from hook_queue_info().
    $info['running'] = isset($info['cron_schedule']) ? 1 : 0;
    return;
  }

  if ($info['run_info'] == 3) {
    // "error encountered while finishing process". (Only for advanced methods.)
    $info['running'] = 2;
  }
  elseif ($info['run_info'] == 1 || $info['run_info'] == 2 || $info['started'] > $info['finished']) {
    // s>f: Job is still running. (We're not verifying by checking queue items.)
    // 1: "running' or "running start; don't touch it".
    // 2: "running but start() should be rerun". (Only for advanced runners.)
    $info['running'] = 3;
  }

  if (isset($info['cron_schedule'])
      && (!isset($info['running']) || $force_calc_next)) {
    // Check time of the next cron run - either for calculating status or
    // because we've been told to.

    // If cron_schedule is zero/empty, the job should always be started, and
    // we can't calculate run_next.
    if (empty($info['started']) || empty($info['cron_schedule'])) {
      if (!isset($info['running'])) {
        $info['running'] = 1;
      }
      return;
    }

    // Check schedule against previous start time. (Not finish time; this way,
    // if e.g. an hourly job once runs for 61 minutes -and cron runs often- it
    // will still be restarted within the same hour, just after it finishes.)
    elseif (is_numeric($info['cron_schedule'])) {
      // Interval is minutes; check against 'last started' timestamp.
      // If cron is run every X minutes and $info['cron_schedule'] is a multiple
      // of that interval, we can miss a cron run if we compare the exact
      // same number of seconds, so substract some from $info['cron_schedule'].
      $info['run_next'] = $info['started'] + $info['cron_schedule'] * 60 - 10;
    }
    else {
      // Check whether this is a valid period string or crontab value. Since
      // these are complicated formats, just feed them through the
      // parsers and see if the return value is valid.
      $crontab = new DrunkinsCronTab($info['cron_schedule']);
      if ($crontab->cron) {
        $info['run_next'] = $crontab->nextTime($info['started']);
      }
      else {
        // date_modify will raise a PHP warning for invalid string. We can't use
        // date_add(..., date_interval_create_from_date_string($info['cron_schedule']))
        // instead; an 'empty' interval will be created without warning for
        // invalid string, but date_add() will yield a wrong date (php bug,
        // apparently) on PHP5.3. PHP5.5 is OK.
        $next = date_modify(date_create('@' . $info['started']), $info['cron_schedule']);
        if ($next) {
          $info['run_next'] = $next->getTimestamp();
        }
      }
      if (!isset($info['run_next'])) {
        watchdog('drunkins', 'Job "@job" cannot be scheduled: interval %i cannot be parsed.',
          array('@job' => $job_id, '%i' => $info['cron_schedule']), WATCHDOG_ERROR);
      }
    }
  }

  if (!isset($info['running'])) {
    $info['running'] =
      isset($info['run_next']) && REQUEST_TIME >= $info['run_next'] ? 1 : 0;
  }
}

/**
 * (Check whether we should) start/finish any jobs due/running on cron.
 *
 * This is usually done on hook_cron(), but certain job/installation specific
 * settings may require some jobs to be handled on hook_cron_queue_info().
 */
function _drunkins_check_cron_jobs($compiling_queue_info = FALSE) {
  if ($compiling_queue_info) {
    return;
/* This is what we documented and should still implement - ALONG with morphing
   'process_items_after_rerun_start' into a reversed capability as documented in
   @todos at drunkins_start_job().

'process_items_after_rerun_start' does not work anymore, at the moment; we don't
care because we do not have any processes that rerun start.

For now, the default is 'process_items_after_rerun_start == FALSE' which means
we don't have elysia2.1 problems by default and we can just return on $compiling_queue_info.
If we make it a reversed capability we will have elysia2.1 problems, so take
care of that. (Make a module-wide setting, or just check elysia version?)


I drafted the following desired behavior, when trying to understand things:

 * When to call what:*

start / enqueue-items-restart:
 - if 'process_items_after_rerun_start' is not set for a job:
   on cron; return items normally in queue-info.
 - if 'process_items_after_rerun_start' is set AND NOT elysia2.1-compat:
   on cron; return items with 'skip on cron' in queue-info.
 - if 'process_items_after_rerun_start' is set AND elysia2.1-compat:
   on cron_queue_info; return queue-info depending on whether process_more is set.
     ^^ <== WE CANNOT do enqueue-items-restart then. If we do that, we might run it twice. Except if we make a new status 'starting' that we alternate between 'starting' and 'running' when running start() / waiting for the next cron run.
        (which is similar to the reason why we have commented out check-finish for now.)
@todo it probably makes sense to have 'process_items_after_rerun_start'=TRUE be
  the default (if this thing gets to be a 'capability'), but then simple processes
  running start() will suffer too. We should move all this into a child runner class.

check-finish:
 - if 'process_items_after_rerun_start' is not set for a job:
   on cron. (We can always just start processing items.)
   (We could maybe make a setting to do it on queue-info but I don't see why.)
 - if 'process_items_after_rerun_start' is set:
   - on cron_queue_info (because we don't know yet whether we can return queue-info)
   - OR on cron but prevent calling start() afterwards so that no queue items will
     be processed during this cron run
     ...depending on runner-ish setting 'prevent_cron_restart_delay'.

 * current situation: *

*for now*, rerunning start *probably* works with us doing everything on cron,
given that the default 'process_items_after_rerun_start' is FALSE at the moment.
DO NOT set it to true, because those jobs will not start. (drunkins_cron_queue_info()
does not return queue info for those processes, since our rewrite to accommodate
elysia 2.1.)


as long as you never set 'process_items_after_rerun_start' to FALSE.

@todo I only briefly checked but it seems that elysia_cron dev > 2.1
  calls this hook a lot. Do we want to cut down on the numberOfItems()
  calls in that case? Should we make a global option that ignores any
  'prevent_cron_restart_delay'? (Maybe even make that the default, and note at
  'prevent_cron_restart_delay' that you can only use this if you set a global variable?)
 */
  }

  $reset_memory_limit = FALSE;

  // Set $job_info by reference because we need the up to date info if it gets
  // changed in _drunkins_queue_process_check_finish().
  foreach (drunkins_job_status_info() as $job_id => &$job_info) {

    if ($job_info['running'] == 3 && $job_info['method'] === 'queue') {
      if (!empty($job_info['process_context']['drunkins_process_more'])) {
        // The job's start() should be called again to fetch / enqueue more
        // items.
        // Please note: the only thing that prevents multiple start() calls
        // (which can take a long time) from overlapping in time, is the cron
        // lock. This module does not do any locking by itself.
//@todo solve this by alternating between a new 'starting' state (which only is set
// during start() call) and 'running' state (which is set while waiting for the next start() call)?
// If we don't do this, we certainly won't be able to call this from hook_cron_queue_info().
        _drunkins_enqueue_items($job_id);
      }
      else {
        // If a job is about to finish (i.e. has 0 items left) and can be
        // restarted on this same cron run, its queue-info must not have a
        // worker callback if the 'process_items_after_rerun_start' runner
        // setting is true. This is why, in that case,
        // - we must either call _drunkins_queue_process_check_finish() *before*
        //   determining the exact queue-info to return on hook_cron_queue_info,
        // - or we must call it with 'job_may_restart_immediately' set to FALSE.
        //
        // Doing the former means:
        // - we can still restart an already-finished job in the same cron run.
        //   (The latter introduces a delay of 1 cron run, so is not good for
        //   jobs that run very often - unless they have almost no chance of
        //   getting into an 'unfinished with 0 items to process' state.)
        // - we make superfluous Queue::numberOfItems() calls (to doublecheck if
        //   a process marked as 'running' still has items) on almost every
        //   hook_cron_queue_info() invocation.
        // - hook_cron_queue_info() could do a lot of processing (if a process
        //   actually gets finished).
        // - unfinished jobs with 0 items are finished also outside of cron runs
        //   when hook_cron_queue_info() is invoked, e.g. when someone views the
        //   job queue UI overview. (This might be a good thing; potentially
        //   confusing for a developer but less confusing for most users.)

        // @todo maybe now that we are clear what needs to happen (see temporary blurb at start of this function),
        //   the above comment is a bit much (because the former is not the default /
        //   is hardly ever done) and can be reworded/moved elsewhere?

        // (In some -dev version between 2015 and 20160625, we chose the former.
        // This introduced a bug, however - while a PHP process was running
        // start(), some other PHP process could execute hook_cron_queue_info()
        // (when viewing the drunkins UI) and 'finish' the starting process,
        // thereby breaking it.
        // So we moved it back to hook_cron(); the 'race condition' does not
        // exist there because Drupal takes care that only one PHP process runs
        // cron.

        //@todo probably, for 'process_items_after_rerun_start' (which will probably be the new default):
        // - add job_may_restart_immediately=FALSE
        // - check this in the start call just below, so we don't call it if we shouldn't.
        // (We don't need to do that now, because 'process_items_after_rerun_start'
        //  does not work / processes never get started. And the whole thing
        //  probably will get moved to a subclass. So we don't bother for now.)

      _drunkins_queue_process_check_finish($job_id, array('job_may_restart_immediately' => TRUE));
      }
    }

    if ($job_info['running'] == 1) {
      // No process is running but it should be.
      drunkins_start_job($job_id, 'queue');
    }
  }

  if ($reset_memory_limit) {
    // If any 'memory_limit_start' was (probably) set in a
    // _drunkins_enqueue_items() call and not reset, reset it here. If
    // it wasn't, this call does nothing / reapplies current limit.
    _drunkins_set_memory_limit('-');
  }
}
